// Base64.cpp: implementation of the CBase64 class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "HTTPServer.h"
#include "Base64.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

// NOTE: Leave the following definitions in the cpp file in
// order to make it private.

#define SKIP (-1)
#define FAIL (-2)
#define PAD  (-3)

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

const char CBase64::szBase64EncodeChars[65] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

const short CBase64::BASE64_DECODE_TABLE[256] = 
{
	FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,SKIP,SKIP,FAIL,FAIL,SKIP,FAIL,FAIL,	/* 0 */
	FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,	/* 1 */
	SKIP,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,0x3e,FAIL,FAIL,FAIL,0x3f,	/* 2 */
	0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,FAIL,FAIL,FAIL,PAD ,FAIL,FAIL,	/* 3 */
	FAIL,0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,	/* 4 */
	0x0f,0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,FAIL,FAIL,FAIL,FAIL,FAIL,	/* 5 */
	FAIL,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f,0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,	/* 6 */
	0x29,0x2a,0x2b,0x2c,0x2d,0x2e,0x2f,0x30,0x31,0x32,0x33,FAIL,FAIL,FAIL,FAIL,FAIL,	/* 7 */
	FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,	/* 8 */
	FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,	/* 9 */
	FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,	/* A */
	FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,	/* B */
	FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,	/* C */
	FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,	/* D */
	FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,	/* E */
	FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,FAIL,	/* F */
  /* 0    1    2    3    4    5    6    7    8    9    A    B    C    D    E    F  */
};

CBase64::CBase64()
{

}

CBase64::~CBase64()
{

}

bool CBase64::Encode(BYTE *szBase64, int iSize, const BYTE *szOriginal)
{
	int i = 0;
	int j = 0;
	UINT iVal = 0;
	int iLength = strlen( (const char *)szOriginal );

	ASSERT( szBase64 != NULL && szOriginal != NULL );
	for( i = 0, j = 0; i + (3 - 1) < iLength && j + (4 - 1) < iSize; i += 3, j += 4 ) {
		iVal = (szOriginal[i] & 0xFC) >> 2;
		ASSERT( 0 <= iVal && iVal < 64 );
		szBase64[j] = szBase64EncodeChars[iVal];

		iVal = ((szOriginal[i] & 0x03) << 4) |
			((szOriginal[i + 1] & 0xF0) >> 4);
		ASSERT( 0 <= iVal && iVal < 64 );
		szBase64[j + 1] = szBase64EncodeChars[iVal];

		iVal = ((szOriginal[i + 1] & 0x0F) << 2) |
			((szOriginal[i + 2] & 0xC0) >> 6);
		ASSERT( 0 <= iVal && iVal < 64 );
		szBase64[j + 2] = szBase64EncodeChars[iVal];

		iVal = szOriginal[i + 2] & 0x3F;
		ASSERT( 0 <= iVal && iVal < 64 );
		szBase64[j + 3] = szBase64EncodeChars[iVal];
	}
	switch( iLength - i ) {
	case 0:
		if( j >= iSize ) {
			return false;
		}
		szBase64[j] = '\0';
		break;
	case 1:
		if( j + 4 >= iSize ) {
			return false;
		}
		szBase64[j + 4] = '\0';
		szBase64[j + 3] = '=';
		szBase64[j + 2] = '=';
		iVal = (szOriginal[i] & 0xFC) >> 2;
		ASSERT( 0 <= iVal && iVal < 64 );
		szBase64[j] = szBase64EncodeChars[iVal];
		iVal = (szOriginal[i] & 0x03) << 4;
		ASSERT( 0 <= iVal && iVal < 64 );
		szBase64[j + 1] = szBase64EncodeChars[iVal];
		break;
	case 2:
		if( j + 4 >= iSize ) {
			return false;
		}
		szBase64[j + 4] = '\0';
		szBase64[j + 3] = '=';
		iVal = (szOriginal[i] & 0xFC) >> 2;
		ASSERT( 0 <= iVal && iVal < 64 );
		szBase64[j] = szBase64EncodeChars[iVal];
		iVal = (szOriginal[i] & 0x03) << 4 |
			(szOriginal[i + 1] & 0xF0) >> 4;
		ASSERT( 0 <= iVal && iVal < 64 );
		szBase64[j + 1] = szBase64EncodeChars[iVal];
		iVal = (szOriginal[i + 1] & 0x0F) << 2;
		ASSERT( 0 <= iVal && iVal < 64 );
		szBase64[j + 2] = szBase64EncodeChars[iVal];
		break;
	default:
		ASSERT( false );
		break;
	}
	return true;
}

bool CBase64::Decode(BYTE *szOriginal, int iSize, const BYTE *szBase64)
{
	int i = 0;
	int j = 0;
	UINT iVal = 0;
	int iLength = strlen( (const char *)szBase64 );

	ASSERT( szOriginal != NULL && iSize >= 0 && szBase64 != NULL );
	if( iLength % 4 != 0 ) {
		return false;
	}
	for( i = 0, j = 0; i + (4 - 1) < iLength && j + (3 - 1) < iSize; i += 4, j += 3 ) {
		memset( &szOriginal[j], 0, 3 );
		iVal = BASE64_DECODE_TABLE[szBase64[i]];
		if( iVal == FAIL || iVal == SKIP ) {
			return false;
		} else if( iVal == PAD ) {
			iVal = 0;
		}
		szOriginal[j] |= iVal << 2;
		iVal = BASE64_DECODE_TABLE[szBase64[i + 1]];
		if( iVal == FAIL || iVal == SKIP ) {
			return false;
		} else if( iVal == PAD ) {
			iVal = 0;
		}
		szOriginal[j] |= iVal >> 4;
		szOriginal[j + 1] |= iVal << 4;
		iVal = BASE64_DECODE_TABLE[szBase64[i + 2]];
		if( iVal == FAIL || iVal == SKIP ) {
			return false;
		} else if( iVal == PAD ) {
			iVal = 0;
		}
		szOriginal[j + 1] |= iVal >> 2;
		szOriginal[j + 2] |= iVal << 6;
		iVal = BASE64_DECODE_TABLE[szBase64[i + 3]];
		if( iVal == FAIL || iVal == SKIP ) {
			return false;
		} else if( iVal == PAD ) {
			iVal = 0;
		}
		szOriginal[j + 2] |= iVal;
	}
	if( j >= iSize ) {
		return false;
	}
	szOriginal[j] = '\0';
	return true;
}
