// CmdLineA.cpp: implementation of the CCmdLineA class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "HTTPServer.h"
#include "CmdLineA.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CCmdLineA::CCmdLineA()
{
	iSwitchCount = 0;
	iGenStrCount = 0;
}

CCmdLineA::~CCmdLineA()
{
}

bool CCmdLineA::AddSwitch(LPCTSTR lpSwitch)
{
	ASSERT(lpSwitch != NULL);
	if (iSwitchCount == MAX_SWITCH_COUNT) {
		return false;
	}
	sarrSwitch[iSwitchCount] = lpSwitch;
	farrSwitch[iSwitchCount] = false;
	iSwitchCount++;
	return true;
}

bool CCmdLineA::Analyze(LPCTSTR lpCmdLine)
{
	enum SwitchScanConstants swtScan;
	TCHAR szCurSwt[MAX_SWITCH_SIZE]; // Current switch
	TCHAR szCurStr[MAX_CMDLNSTR_SIZE]; // Current string
	const TCHAR *pchSrc = NULL;
	bool fInQuote = false;
	int iSubDst = 0; // Subscript for the destination string
	enum SwitchScanConstants swtScan2;
	ASSERT(lpCmdLine != NULL);

	GetReadyForAnalysis();
	swtScan = SS_GENSTR;
	szCurSwt[0] = _T('\0');
	szCurStr[0] = _T('\0');
	pchSrc = lpCmdLine;
	fInQuote = false;
	iSubDst = 0;
	while (*pchSrc) {
		// Save a copy of the state before it changes.
		swtScan2 = swtScan;
		switch (*pchSrc) {
		case _T('/'):
			// If the position is not in a quoted string, turn the scan indicator
			// into SS_SWITCH.
			if (!fInQuote) {
				swtScan = SS_SWITCH;
			} // Else the string in quote cannot be a switch.
			// NOTE: according to the way to handle the case here, 
			// any file name as a parameter that contains a slash
			// '/' should be quoted on the command line.
			break;
		case _T('-'):
			if (!fInQuote) {
				if (pchSrc != lpCmdLine) {
					// Current position is not at head of the command line.
					// See the character on the left of current position.
					if (*_tcsdec(lpCmdLine, pchSrc) == ' ') {
						// The character on the left of current position is
						// ' ', so turn the scan indicator into SS_SWITCH.
						swtScan = SS_SWITCH;
					} // Else the '-' is a part of a string so don't change the
						// state indicator.
				} else {
					// Current position is at the head of the command line.
					// Just turn the scan indicator into SS_SWITCH.
					swtScan = SS_SWITCH;
				}
			} // Else the string in quote cannot be a switch.
			// NOTE: according to the way to handle the case here, 
			// any file name as a parameter that contains a hyphen
			// '-' needn't be quoted on the command line.
			break;
		case _T(' '):
			if (!fInQuote) {
				if (swtScan != SS_SWITCH) { // It is not right after a switch.
					swtScan = SS_GENSTR; // The new string is general.
				} else {
					swtScan = SS_SWTSTR; // The new string is a switch setting.
				}
			} // Else don't change it.
			break;
		case _T('\"'):
			if (swtScan == SS_SWITCH) {
				swtScan = SS_SWTSTR;
			} // Else swtScan is SS_SWTSTR or SS_GENSTR so don't change it.
			fInQuote = !fInQuote;
			break;
		default:
			break;
		}
		if (swtScan2 != swtScan) {
			// The state has changed.
			// Save string information if necessary.
			// See which state change has happened.
			if (swtScan2 == SS_SWTSTR || (swtScan == SS_GENSTR && swtScan2 == SS_SWITCH)) {
				// Scan of a switch is over.
				if (szCurSwt[0]) {
					// A switch is specified.
					// Save the string with the corresponding switch.
					{
						int iCorSwt = 0; // Corresponding switch
						// Complete the switch if necessary.
						if (swtScan2 == SS_SWITCH) {
							szCurSwt[iSubDst] = _T('\0');
						}
						for (iCorSwt = 0; iCorSwt < iSwitchCount; iCorSwt++) {
							if (sarrSwitch[iCorSwt].CompareNoCase(szCurSwt) == 0) {
								break;
							}
						}
						if (iCorSwt == iSwitchCount) {
							// Unexpected switch
							return false;
						}
						farrSwitch[iCorSwt] = true;
						// Complete the string before saving.
						if (swtScan2 == SS_SWTSTR) {
							// The previous state is SS_SWTSTR.
							szCurStr[iSubDst] = _T('\0');
							sarrSetting[iCorSwt] = szCurStr;
						} else {
							// The previous state is SS_SWITCH.
							// No switch setting is found.
							sarrSetting[iCorSwt] = _T("");
						}
					}
				} else {
					// There is no switch specified.
					// It is a format error of the command line.
					return false;
				}
			} // End if for saving switch information.
			if (swtScan2 == SS_GENSTR) {
				// Save the general string if necessary.
				if (szCurStr[0]) {
					szCurStr[iSubDst] = _T('\0');
					if (iGenStrCount < MAX_SWITCH_COUNT) {
						sarrGenStr[iGenStrCount] = szCurStr;
						iGenStrCount++;
					} else {
						// Overflowed.
						return false;
					}
				} // Else the general string is empty so do nothing.
			} // End if for saving general string information.
			if (swtScan == SS_SWITCH) {
				// Get ready to retrieve the new switch.
				szCurSwt[0] = _T('\0');
				iSubDst = 0;
			} else { // SS_GENSTR or SS_SWTSTR
				// The status changed from switch to string.
				// Get ready to retrieve the string.
				// Complete current switch before going on.
				if (swtScan2 == SS_SWITCH) {
					szCurSwt[iSubDst] = _T('\0');
				} // Else the string must have been retrieved.
				szCurStr[0] = _T('\0');
				iSubDst = 0;
			}
		}
		switch (swtScan) {
		case SS_GENSTR:
			if ((!fInQuote && *pchSrc != _T('\"') && *pchSrc != _T(' ')) ||
				(fInQuote && *pchSrc != _T('\"'))) {
				if (_tcsinc(pchSrc) - pchSrc == 2) {
					if (iSubDst + 2 < MAX_CMDLNSTR_SIZE) {
						szCurStr[iSubDst] = *pchSrc;
						szCurStr[iSubDst + 1] = *(pchSrc + 1);
						iSubDst += 2;
					} // Else the string is too long but ignore the error.
				} else {
					if (iSubDst + 1 < MAX_CMDLNSTR_SIZE) {
						szCurStr[iSubDst] = *pchSrc;
						iSubDst++;
					} // Else the string is too long but ignore the error.
				}
			} else {
				// A state change is met. Save current string and get ready
				// for the next.
				if (szCurStr[0]) {
					// Complete the string before saving.
					szCurStr[iSubDst] = _T('\0');
					if (iGenStrCount < MAX_SWITCH_COUNT) {
						sarrGenStr[iGenStrCount] = szCurStr;
						iGenStrCount++;
					} else {
						// Overflowed.
						return false;
					}
				} // Else do nothing.
				// Reset the string.
				szCurStr[0] = _T('\0');
				iSubDst = 0;
			}
			break;
		case SS_SWITCH:
			if (*pchSrc != _T('/') && *pchSrc != _T('-')) {
				if (_tcsinc(pchSrc) - pchSrc == 2) {
					if (iSubDst + 2 < MAX_SWITCH_SIZE) {
						szCurSwt[iSubDst] = *pchSrc;
						szCurSwt[iSubDst + 1] = *(pchSrc + 1);
						iSubDst += 2;
					} // Else do nothing.
				} else {
					if (iSubDst + 1 < MAX_SWITCH_SIZE) {
						szCurSwt[iSubDst] = *pchSrc;
						iSubDst++;
					} // Else do nothing.
				}
			} // Else don't treat it.
			// NOTE: A switch cannot contain a slash ('/') or a hyphen ('-').
			break;
		case SS_SWTSTR:
			if ((!fInQuote && *pchSrc != _T('\"') && *pchSrc != _T(' ')) ||
				(fInQuote && *pchSrc != _T('\"'))) {
				if (_tcsinc(pchSrc) - pchSrc == 2) {
					if (iSubDst + 2 < MAX_CMDLNSTR_SIZE) {
						szCurStr[iSubDst] = *pchSrc;
						szCurStr[iSubDst + 1] = *(pchSrc + 1);
						iSubDst += 2;
					} // Else do nothing.
				} else {
					if (iSubDst + 1 < MAX_CMDLNSTR_SIZE) {
						szCurStr[iSubDst] = *pchSrc;
						iSubDst++;
					} // Else do nothing.
				}
			} // Else don't treat it.
			break;
		}
		pchSrc = _tcsinc(pchSrc);
	}
	// Save the last part.
	if (swtScan == SS_SWITCH || swtScan == SS_SWTSTR) {
		{
			int iCorSwt = 0; // Corresponding switch
			// Complete the switch if necessary.
			if (swtScan == SS_SWITCH) { // Use swtScan but not swtScan2 because 
				// swtScan is not changed.
				szCurSwt[iSubDst] = _T('\0');
			}
			for (iCorSwt = 0; iCorSwt < iSwitchCount; iCorSwt++) {
				if (sarrSwitch[iCorSwt].CompareNoCase(szCurSwt) == 0) {
					break;
				}
			}
			if (iCorSwt == iSwitchCount) {
				// Unexpected switch
				return false;
			}
			farrSwitch[iCorSwt] = true;
			// Complete the string before saving.
			if (swtScan == SS_SWTSTR) {
				// It is in the switch string state so complete the string
				// and add it as a switch setting.
				szCurStr[iSubDst] = _T('\0');
				sarrSetting[iCorSwt] = szCurStr;
			} else {
				// The switch is the only part retrieved.
				// Save an empty string as the switch setting.
				sarrSetting[iCorSwt] = _T("");
			}
		}
	} else { // SS_GENSTR
		if (szCurStr[0]) {
			szCurStr[iSubDst] = _T('\0');
			if (iGenStrCount < MAX_SWITCH_COUNT) {
				sarrGenStr[iGenStrCount] = szCurStr;
				iGenStrCount++;
			} else {
				return false;
			}
		}
	}
	return true;
}

void CCmdLineA::Clear()
{
	iSwitchCount = 0;
	iGenStrCount = 0;
}

void CCmdLineA::GetReadyForAnalysis()
{
	int i;
	for (i = 0; i < iSwitchCount; i++) {
		farrSwitch[i] = false;
	}
	iGenStrCount = 0;
}

#ifdef _DEBUG
void CCmdLineA::DebugTest()
{
	AddSwitch(_T("print"));
	AddSwitch(_T("play"));
	AddSwitch(_T("rootdir"));
	AddSwitch(_T("path"));
	Analyze(_T(
		"/print \"file print\" filea \"filea b\" /play fileplay fileb /rootdir rootdir"
		"-path path-a"
	));
	TRACE(_T("End of debug test.\n"));
}
#endif
