// ServerThread.cpp: implementation of the ServerThread class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "HTTPServer.h"
#include "ServerThread.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

/**
 * INFORMATION RECEIVING MODEL
 * 10 Get information from the socket buffer.
 * 20 Append information to the request buffer.
 * 30 Treat the request until there is not a complete line. 
 *    If the buffer is still full, send an error message.
 * 40 Save current state and goto 10.
 */
/**
 * THE LISTEN AND STOP MODEL
 * When the user presses the Listen button, the server socket is 
 * generated and passed to the newly started thread.
 * When the user presses the Stop button, the server socket is 
 * closed and further connection cannot be accepted.
 * After all threads having exited, the server socket is closed,
 * and the server is again ready for listening.
 * Some threads may not stop until something more is 
 * received. When they receive some more information, they see
 * the flag toStop and then quit.
 */
// NOTE: For future modifications, please re-check and rewrite comments.

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CServerThread::CServerThread()
{

}

CServerThread::~CServerThread()
{

}

const int CServerThread::s_iTimeoutLimit = 15;

UINT CServerThread::EntryProcedure(LPVOID lpParam)
{
	STHREAD_INFO *pSThreadInfo = (STHREAD_INFO *)lpParam;
	sockaddr_in isaClient; // Internet client socket address
	int iISACLength = sizeof( isaClient );
	SOCKET socketClient;

	// Increase the thread count.
	pSThreadInfo->csMain.Lock();
	pSThreadInfo->iThreadCount++;
	pSThreadInfo->csMain.Unlock();
	// Ready to treat requests.
	fd_set fdsOne;
	timeval tvOne = { s_iTimeoutLimit, 0 };
	int iBytesReceived = 0;
	// The receive buffer is not a zero-terminated string, but
	// a character array.
	char sReceiveBuffer[SRECEIVEBUFFER_SIZE] = "";
	// The original length of the receive buffer
	// Useful when the request buffer is full: when the request buffer is
	// full, exceeding information can be stored in the receive buffer.
	// The variable indicates the length of the information before
	// receiving.
	int iRcvOLength = 0;
	// The request buffer. It should be able to hold a line of request.
	// The array is used as a queue. iRBIBegin and iRBIEnd indicate the
	// begin and end (over the last effective byte) of the buffer.
	char *sRequestBuffer = new char[SREQUESTBUFFER_SIZE];
	int iRBIBegin = 0; // Request buffer information begin
	int iRBIEnd = 0; // Request buffer information end
	int iResponseOne = 0;
	bool fBlockReception = false;
	// Treat request state. Used when treating a request.
	struct tagTreatRequestState trsOne = {
		INVALID_SOCKET,
		pSThreadInfo->socketServer,
		trHead,
		gtURIProtocol,
		rmUnknown,
		NULL,
		NULL,
		NULL,
		(LPCTSTR)pSThreadInfo->sBasePath,
		false,
		false,
		false,
		"",
		NULL,
		"",
		"",
		pSThreadInfo->sarrPasswords,
		pSThreadInfo->iPasswordCount,
		pSThreadInfo->plistMonitor,
		(LPCTSTR)pSThreadInfo->sDefaultPage,
		pSThreadInfo->pfRecToMon,
		&(pSThreadInfo->csMain),
		pSThreadInfo->hwndDialog
	};

	// Allocate dynamic memory space and clear strings/arrays.
	// NOTE: If memory space expected is equal to or larger than
	// 1024 bytes, I use dynamic allocation. If not, I may use either
	// dynamic allocation or stack allocation, depending on
	// whether the function would be deeply recursed.
	trsOne.szURI = new char[MAX_URI];
	trsOne.szURI[0] = '\0';
	trsOne.szPath = new char[MAX_URI];
	trsOne.szPath[0] = '\0';
	trsOne.sSendBuffer = new char[SENDBUFFER_SIZE];
	memset( trsOne.sSendBuffer, 0, SENDBUFFER_SIZE );
	trsOne.szFullPath = new char[MAX_URI];
	trsOne.szFullPath[0] = '\0';
	TRACE( _T( "Memory allocated\n" ) );

	// Try to accept a connection.
	if(
		(socketClient = accept(
			pSThreadInfo->socketServer,
			(SOCKADDR *)&isaClient,
			&iISACLength
		)) == INVALID_SOCKET
	) {
		TRACE ( _T( "Acception failed.\n" ) );
		goto labelAfterReceiveLoop; // skip treating request if acception failed.
	}
	// Start another thread to accept further connections.
	// After treating requests, let this thread exit and let 
	// new threads treat pending requests.
	// NOTE: Information in the STHREAD_INFO structure is shared.
	// Any thread who wants to modify it should lock the
	pSThreadInfo->csMain.Lock();
	if (!pSThreadInfo->fToStop) {
		AfxBeginThread( EntryProcedure, pSThreadInfo );
	}
	pSThreadInfo->csMain.Unlock();
	// Initialize variables.
	memset( sReceiveBuffer, 0, SRECEIVEBUFFER_SIZE );
	iRcvOLength = 0;
	memset( sRequestBuffer, 0, SREQUESTBUFFER_SIZE );
	iRBIBegin = 0;
	iRBIEnd = 0;
	trsOne.socketClient = socketClient;
	trsOne.state = trHead;
	trsOne.getState = gtURIProtocol;
	trsOne.iRMethod = rmUnknown;
	trsOne.szURI[0] = '\0';
	trsOne.szPath[0] = '\0';
	memset( trsOne.sSendBuffer, 0, SENDBUFFER_SIZE );
	trsOne.fKeepAlive = false;
	trsOne.fRequestHead = false;
	trsOne.fIMS = false;
	trsOne.szIMS[0] = '\0';
	trsOne.szFullPath[0] = '\0';
	trsOne.szUserName[0] = '\0';
	trsOne.szPassword[0] = '\0';
	TRACE( _T( "Accepted. Socket address family: %hd.\n"
		"Socket port: %hu. Socket address: %d.%d.%d.%d.\n"
		"Socket address length: %d.\n" ),
		isaClient.sin_family,
		ntohs( isaClient.sin_port ),
		(int)(isaClient.sin_addr.S_un.S_un_b.s_b1),
		(int)(isaClient.sin_addr.S_un.S_un_b.s_b2),
		(int)(isaClient.sin_addr.S_un.S_un_b.s_b3),
		(int)(isaClient.sin_addr.S_un.S_un_b.s_b4),
		iISACLength
	);
	{
		// Send monitor information.
		char *pszMonitor = new char[80];
		wsprintf( pszMonitor,
			_T( "Accepted connection from address %d.%d.%d.%d, "
			"port number %hu." ),
			(int)(isaClient.sin_addr.S_un.S_un_b.s_b1),
			(int)(isaClient.sin_addr.S_un.S_un_b.s_b2),
			(int)(isaClient.sin_addr.S_un.S_un_b.s_b3),
			(int)(isaClient.sin_addr.S_un.S_un_b.s_b4),
			ntohs( isaClient.sin_port )
		);
		SendMessage(pSThreadInfo->hwndDialog, MYWM_MONITOR_ADD_STRING, (WPARAM)0, (LPARAM)pszMonitor);
		delete pszMonitor;
	}

	// Test the client socket.
	fdsOne.fd_count = 1;
	fdsOne.fd_array[0] = socketClient;
	iResponseOne = select( 0, &fdsOne, NULL, NULL, &tvOne );
	switch( iResponseOne ) {
	case 0:
		TRACE( _T( "Client socket timed out.\n" ) );
		break;
	case SOCKET_ERROR:
		TRACE( _T( "Client socket error.\n" ) );
		break;
	case 1:
		break;
	default:
		ASSERT( false );
		TRACE( _T( "Unexpected response.\n" ) );
		break;
	}
	if( iResponseOne != 1 ) {
		// Force a shutdown.
		ShutSocketDown( socketClient );
		closesocket( socketClient );
		goto labelAfterReceiveLoop;
	}
	// Receive requests and information.
	while( true ) {
		// Receive information. The reception may be blocked when a request
		// is still under process, or the request buffer is full (possibly
		// due to a very long line).
		if( !fBlockReception ) {
			// Receive some thing.
			// NOTE FOR THE recv FUNCTION:
			// In MSDN, there are two constants for the flags
			// parameter: MSG_PEEK and MSG_OOB. Actually, 0
			// is the basic constant.
			iBytesReceived = recv( socketClient, 
				sReceiveBuffer + iRcvOLength, // Receive at the position 
					//sReceiveBuffer + iRcvOLength.
				SRECEIVEBUFFER_SIZE - iRcvOLength, // Maximum number of receivable bytes
				0 // Flags
			); 
			switch( iBytesReceived ) {
			case 0:
				TRACE( _T( "Disconnected. iBytesReceived: 0.\n" ) );
				ShutSocketDown( pSThreadInfo->socketServer );
				if( closesocket( socketClient ) == SOCKET_ERROR ) {
					TRACE( _T( "Failed to close the client socket.\n" ) );
				}
				goto labelAfterReceiveLoop;
				break;
			case SOCKET_ERROR:
				TRACE( _T( "Disconnected. iBytesReceived: SOCKET_ERROR.\n" ) );
				ShutSocketDown( pSThreadInfo->socketServer );
				if( closesocket( socketClient ) == SOCKET_ERROR ) {
					TRACE( _T( "Failed to close the client socket.\n" ) );
				}
				goto labelAfterReceiveLoop;
				break;
			default:
				// Successfully received information.
				TRACE(_T("Bytes received: %d.\n"), iBytesReceived);
#ifdef _DEBUG
				{
					char szDisplay[SRECEIVEBUFFER_SIZE + 1] = "";
					int i = iRcvOLength;
					for( i = iRcvOLength; i < iBytesReceived; i++ ) {
						szDisplay[i] = sReceiveBuffer[i];
					}
					szDisplay[i] = '\0';
					TRACE(_T("%s\n"), &szDisplay[iRcvOLength]);
				}
#endif
				break;
			}
		}
		// Fill the request buffer with information in the
		// receive buffer. If there is not enough space, block
		// reception until there is space freed from the
		// request buffer.
		{
			int i;
			int j;

			fBlockReception = false;
			// j is a subscript for sRequestBuffer.
			// i is a subscript for sReceiveBuffer.
			// Reaching the end of sReceiveBuffer is as the break condition 
			// of the for loop.
			for( i = 0, j = iRBIEnd; i < iBytesReceived; i++, j++ ) {
				// Manually round j up when j reaches the end of sRequestBuffer.
				if( j == SREQUESTBUFFER_SIZE ) {
					j = 0;
				}
				// Manually break when sRequestBuffer is full, and set
				// fBlockReception to true.
				if( iRBIBegin == 0 ) {
					if( j == SREQUESTBUFFER_SIZE - 1 ) {
						fBlockReception = true;
						break;
					}
				} else {
					if( j == iRBIBegin - 1 ) {
						fBlockReception = true;
						break;
					}
				}
				// Copy one byte.
				sRequestBuffer[j] = sReceiveBuffer[i];
			}
			// Manually round j up (after the last j++).
			// If the loop is broken due to sRequestBuffer being full,
			// j is already in the valid range.
			if( j == SREQUESTBUFFER_SIZE ) {
				j = 0;
			}
			// Set the new iRBIEnd.
			iRBIEnd = j;
			// fBlockReception being true equals to sRequestBuffer being full.
			if( fBlockReception ) {
				// If sRequestBuffer is full, move unbuffered data 
				// from the end of the receive buffer to the 
				// beginning.
				// Currently i is pointing to the first byte that
				// hasn't been sent into the request buffer.
				if( i > 0 ) {
					// i > 0 means some information has been copied into
					// the request buffer.
					for( j = 0; i < iBytesReceived; i++, j++ ) {
						sReceiveBuffer[j] = sReceiveBuffer[i];
					}
					iRcvOLength = j;
				} else {
					// The buffer is not modified, so data needn't be moved.
					iRcvOLength = iBytesReceived;
				}
			} else {
				// All information has been sent. Reset iRcvOLength.
				iRcvOLength = 0;
			}
		}
#ifdef _DEBUG
		ShowRequest( sRequestBuffer, iRBIBegin, iRBIEnd );
#endif
		TRACE( _T( "Remaining buffer information: \n%s\n" ), sReceiveBuffer );
		// Treat the request until it cannot be treated.
		while( (iResponseOne = TreatRequest( &trsOne, sRequestBuffer, &iRBIBegin, iRBIEnd )) != RQ_INCOMPLETE ) {
			if( iResponseOne == RQ_ERROR ) {
				// Bad Request
				SendHTTPError( 400, socketClient );
				// Force a shut down.
				ShutSocketDown( socketClient );
				goto labelAfterReceiveLoop;
			} else if( iResponseOne == RQ_SERVER_ERROR ) {
				// Internal Server Error
				SendHTTPError( 500, socketClient );
				ShutSocketDown( socketClient );
				goto labelAfterReceiveLoop;
			}
		}
		// Check whether the user interface has closed the server socket.
		pSThreadInfo->csMain.Lock();
		if( pSThreadInfo->fToStop ) {
			pSThreadInfo->csMain.Unlock();
			// Force a shut down.
			ShutSocketDown( socketClient );
			closesocket( socketClient );
			break;
		}
		pSThreadInfo->csMain.Unlock();
	}
labelAfterReceiveLoop:
	// Free up dynamically allocated spaces.
	delete [] trsOne.szURI;
	trsOne.szURI = NULL;
	delete [] trsOne.szPath;
	trsOne.szPath = NULL;
	delete [] trsOne.sSendBuffer;
	trsOne.sSendBuffer = NULL;
	delete [] trsOne.szFullPath;
	trsOne.szFullPath = NULL;
	delete [] sRequestBuffer;
	sRequestBuffer = NULL;
	TRACE( _T( "Memory released.\n" ) );
	// Decrease the thread count and notify the main window.
	pSThreadInfo->csMain.Lock();
	pSThreadInfo->iThreadCount--;
	pSThreadInfo->csMain.Unlock();
	::PostMessage( pSThreadInfo->hwndDialog, WM_THREAD_FINISHED, 0, 0 );
	return 0;
}

// CServerThread::TreatRequest
// This function treats requests.
// pTRState - current state treatment.
int CServerThread::TreatRequest(struct tagTreatRequestState *pTRState, const char *sRequest, int *piBegin, int iEnd)
{
	int iResponseOne = RQ_SUCCESSFUL;
	bool fResponseTwo = false;
	ASSERT( pTRState != NULL && sRequest != NULL );

	while( true ) {
		switch( pTRState->state ) {
		case trHead: // Request head ("<METHOD> <URI> HTTP/<HTTP VERSION>" 
			// Return if there is not enough information.
			if( (iResponseOne = TreatRequest_HasALine2( sRequest, piBegin, iEnd )) != RQ_SUCCESSFUL ) {
				return iResponseOne;
			}
			// Ignore blank lines.
			while( ExtractWord( NULL, SREQUESTBUFFER_SIZE + 1, sRequest, *piBegin, iEnd ) == 0 ) {
				if( IgnoreLine( sRequest, piBegin, iEnd ) == RQ_INCOMPLETE ) {
					return RQ_INCOMPLETE;
				}
			}
			// Decide the request method.
			if( (pTRState->iRMethod = AnalyzeRequestMethod(sRequest, piBegin, iEnd)) == rmUnknown ) {
				TRACE( _T( "Unknown or invalid request.\n" ) );
				*piBegin = iEnd;
				return RQ_ERROR;
			} else if( pTRState->iRMethod == rmIncomplete ) {
				return RQ_INCOMPLETE;
			} // Else it is successful.
			// Initializations for different methods are listed here.
			switch( pTRState->iRMethod ) {
			case rmOptions:
				break;
			case rmGet:
				// Maintain these one-time variables.
				pTRState->szURI[0] = '\0';
				pTRState->szPath[0] = '\0';
				pTRState->szFullPath[0] = '\0';
				pTRState->getState = gtURIProtocol;
				pTRState->fRequestHead = false;
				break;
			case rmHead:
				pTRState->szURI[0] = '\0';
				pTRState->szPath[0] = '\0';
				pTRState->szFullPath[0] = '\0';
				pTRState->getState = gtURIProtocol;
				pTRState->fRequestHead = true;
				break;
			case rmPost:
				break;
			case rmPut:
				break;
			case rmDelete:
				break;
			case rmTrace:
				break;
			case rmConnect:
				break;
			default:
				break;
			}
			pTRState->state = trBody;
			break;
		case trBody:
			switch( pTRState->iRMethod ) {
			case rmOptions: // OPTIONS - Currently unimplemented.
				TRACE( _T( "Request type: OPTIONS.\n" ) );
				// Check if there is at least one line.
				if( (iResponseOne = TreatRequest_HasALine2( sRequest, piBegin, iEnd )) != RQ_SUCCESSFUL ) {
					return iResponseOne;
				}
				// Loop until the first empty line is met.
				while( fResponseTwo = TreatRequest_HasALine( sRequest, *piBegin, iEnd ) ) {
					if( ExtractWord( NULL, SREQUESTBUFFER_SIZE + 1, sRequest, *piBegin, iEnd ) == 0 ) {
						// The line is empty.
						break;
					}
					if( IgnoreLine( sRequest, piBegin, iEnd ) == RQ_INCOMPLETE ) {
						ASSERT( false ); // This is impossible.
						return RQ_INCOMPLETE;
					}
				}
				if( !fResponseTwo ) {
					return RQ_INCOMPLETE;
				}
				SendHTTPError( 501, pTRState->socketClient );
				break;
			case rmGet:
				TRACE( _T( "Request type: GET.\n" ) );
				// Treat the request.
				iResponseOne = TreatRequestGet( pTRState, sRequest, piBegin, iEnd );
				if( iResponseOne == RQ_ERROR || iResponseOne == RQ_SERVER_ERROR ) {
					TRACE( _T( "Get request failed.\n" ) );
					// Clear the buffer.
					*piBegin = iEnd;
					// Reset the state.
					pTRState->state = trHead;
					return iResponseOne;
				} else if( iResponseOne == RQ_INCOMPLETE ) {
					return RQ_INCOMPLETE;
				}
				break;
			case rmHead:
				TRACE( _T( "Request type: HEAD.\n" ) );
				iResponseOne = TreatRequestGet( pTRState, sRequest, piBegin, iEnd );
				if( iResponseOne == RQ_ERROR || iResponseOne == RQ_SERVER_ERROR ) {
					TRACE( _T( "Head request failed.\n" ) );
					*piBegin = iEnd;
					pTRState->state = trHead;
					return iResponseOne;
				} else if( iResponseOne == RQ_INCOMPLETE ) {
					return RQ_INCOMPLETE;
				}
				break;
			case rmPost:
				TRACE( _T( "Request type: POST.\n" ) );
				// Doesn't check if there is at least one line
				// because a file may be being sent.
				// Currently not implemented. Directly send the error 
				// message.
				SendHTTPError( 501, pTRState->socketClient );
			case rmPut:
				TRACE( _T( "Request type: PUT.\n" ) );
				if( (iResponseOne = TreatRequest_HasALine2( sRequest, piBegin, iEnd )) != RQ_SUCCESSFUL ) {
					return iResponseOne;
				}
				while( fResponseTwo = TreatRequest_HasALine( sRequest, *piBegin, iEnd ) ) {
					if( ExtractWord( NULL, SREQUESTBUFFER_SIZE + 1, sRequest, *piBegin, iEnd ) == 0 ) {
						break;
					}
					if( IgnoreLine( sRequest, piBegin, iEnd ) == RQ_INCOMPLETE ) {
						ASSERT( false );
						return RQ_INCOMPLETE;
					}
				}
				if( !fResponseTwo ) {
					return RQ_INCOMPLETE;
				}
				SendHTTPError( 501, pTRState->socketClient );
				break;
			case rmDelete:
				TRACE( _T( "Request type: DELETE.\n" ) );
				if( (iResponseOne = TreatRequest_HasALine2( sRequest, piBegin, iEnd )) != RQ_SUCCESSFUL ) {
					return iResponseOne;
				}
				while( fResponseTwo = TreatRequest_HasALine( sRequest, *piBegin, iEnd ) ) {
					if( ExtractWord( NULL, SREQUESTBUFFER_SIZE + 1, sRequest, *piBegin, iEnd ) == 0 ) {
						break;
					}
					if( IgnoreLine( sRequest, piBegin, iEnd ) == RQ_INCOMPLETE ) {
						ASSERT( false );
						return RQ_INCOMPLETE;
					}
				}
				if( !fResponseTwo ) {
					return RQ_INCOMPLETE;
				}
				SendHTTPError( 501, pTRState->socketClient );
				break;
			case rmTrace:
				TRACE( _T( "Request type: TRACE.\n" ) );
				if( (iResponseOne = TreatRequest_HasALine2( sRequest, piBegin, iEnd )) != RQ_SUCCESSFUL ) {
					return iResponseOne;
				}
				while( fResponseTwo = TreatRequest_HasALine( sRequest, *piBegin, iEnd ) ) {
					if( ExtractWord( NULL, SREQUESTBUFFER_SIZE + 1, sRequest, *piBegin, iEnd ) == 0 ) {
						break;
					}
					if( IgnoreLine( sRequest, piBegin, iEnd ) == RQ_INCOMPLETE ) {
						ASSERT( false );
						return RQ_INCOMPLETE;
					}
				}
				if( !fResponseTwo ) {
					return RQ_INCOMPLETE;
				}
				SendHTTPError( 501, pTRState->socketClient );
				break;
			case rmConnect:
				TRACE( _T( "Request type: CONNECT.\n" ) );
				if( (iResponseOne = TreatRequest_HasALine2( sRequest, piBegin, iEnd )) != RQ_SUCCESSFUL ) {
					return iResponseOne;
				}
				while( fResponseTwo = TreatRequest_HasALine( sRequest, *piBegin, iEnd ) ) {
					if( ExtractWord( NULL, SREQUESTBUFFER_SIZE + 1, sRequest, *piBegin, iEnd ) == 0 ) {
						break;
					}
					if( IgnoreLine( sRequest, piBegin, iEnd ) == RQ_INCOMPLETE ) {
						ASSERT( false );
						return RQ_INCOMPLETE;
					}
				}
				if( !fResponseTwo ) {
					// The request is not complete.
					return RQ_INCOMPLETE;
				}
				SendHTTPError( 501, pTRState->socketClient );
				break;
			case rmIncomplete:
				// Blank lines has been passed, so this is impossible.
				ASSERT( false );
				break;
			default:
				TRACE( _T( "Should not return this kind of request.\n" ) );
				ASSERT( false );
				break;
			}
			goto labelAfterRequest;
			break;
		default:
			ASSERT( false );
			break;
		}
	}
labelAfterRequest:
	pTRState->state = trHead;
	*piBegin = iEnd;
	return RQ_SUCCESSFUL;
}

CServerThread::RequestMethodConstants CServerThread::AnalyzeRequestMethod(const char *sRequest, int *piBegin, int iEnd)
{
	char szMethod[32] = "";
	int iBytesExtracted;
	ASSERT( sRequest != NULL && 0 <= *piBegin && *piBegin < SREQUESTBUFFER_SIZE &&
		0 <= iEnd && iEnd < SREQUESTBUFFER_SIZE	);
	iBytesExtracted = ExtractWord( szMethod, 32, sRequest, *piBegin, iEnd );
	if( iBytesExtracted == 0 ) return rmIncomplete; // The line is empty, but it is impossible.
	if( iBytesExtracted < 0 ) { // Word too long
		*piBegin -= iBytesExtracted;
		if( *piBegin >= SREQUESTBUFFER_SIZE ) {
			*piBegin -= SREQUESTBUFFER_SIZE;
		}
		return rmUnknown;
	}
	*piBegin += iBytesExtracted;
	if( *piBegin >= SREQUESTBUFFER_SIZE ) {
		*piBegin -= SREQUESTBUFFER_SIZE;
	}
	if( strcmp( szMethod, "OPTIONS" ) == 0 ) {
		return rmOptions;
	} else if( strcmp( szMethod, "GET" ) == 0 ) {
		return rmGet;
	} else if( strcmp( szMethod, "HEAD" ) == 0 ) {
		return rmHead;
	} else if( strcmp( szMethod, "POST" ) == 0 ) {
		return rmPost;
	} else if( strcmp( szMethod, "PUT" ) == 0 ) {
		return rmPut;
	} else if( strcmp( szMethod, "DELETE" ) == 0 ) {
		return rmDelete;
	} else if( strcmp( szMethod, "TRACE" ) == 0 ) {
		return rmTrace;
	} else if( strcmp( szMethod, "CONNECT" ) == 0 ) {
		return rmConnect;
	} else {
		return rmUnknown;
	}
}

int CServerThread::TreatRequestGet(struct tagTreatRequestState *pTRState, const char *sRequest, int *piBegin, int iEnd)
{
	int iReturnValue = RQ_SUCCESSFUL;
	int iResponseOne = 0;
	const int iBufferLength = 100;
	char szBuffer[iBufferLength] = "";
	ASSERT( pTRState != NULL && sRequest != NULL );

	while( true ) {
		switch( pTRState->getState ) {
		case gtURIProtocol:
			// Ensure there is a complete line.
			if( (iResponseOne = TreatRequest_HasALine2( sRequest, piBegin, iEnd )) != RQ_SUCCESSFUL ) {
				return iResponseOne;
			}
			if( (iResponseOne = ExtractWord( pTRState->szURI, MAX_URI, sRequest, *piBegin, iEnd )) == 0 ) {
				// Failed to extract a word. This means the line is empty.
				pTRState->szURI[0] = '\0'; // The line is empty so the word is empty.
				// This means the request is invalid.
				TRACE( _T( "Bad GET request - no URI specified.\n" ) );
				pTRState->state = trHead;
				return RQ_ERROR;
			} else if( iResponseOne < 0 ) {
				// The word is too long.
				*piBegin -= iResponseOne;
				if( *piBegin >= SREQUESTBUFFER_SIZE ) {
					*piBegin -= SREQUESTBUFFER_SIZE;
				}
				TRACE( _T( "Format error (word too long)!\n" ) );
				pTRState->state = trHead;
				return RQ_ERROR;
			}
			// Ignore rest information on the same line.
			if( (iResponseOne = IgnoreLine( sRequest, piBegin, iEnd )) == RQ_INCOMPLETE ) {
				ASSERT( false );
				return RQ_INCOMPLETE;
			}
			// Store the URI in all URI form, relative path form and full path form.
			TRACE( _T( "URI Received: %s\n" ), (LPCTSTR)pTRState->szURI );
			if( (iResponseOne = ConvertURIToPath( pTRState->szPath, MAX_URI, pTRState->szURI )) != RQ_SUCCESSFUL ) {
				TRACE( _T( "Failed to convert the URI to path.\n" ) );
				pTRState->szPath[0] = '\0';
				pTRState->state = trHead;
				return iResponseOne; // RQ_ERROR or RQ_SERVER_ERROR
			} else {
				TRACE( _T( "Converted to path: %s\n" ), (LPCTSTR)pTRState->szPath );
			}
			// Check if it is the empty (this means the URI is pointing to the root).
			if( pTRState->szPath[0] == '\0' ) {
				// Make it point to the default page.
				strncpy( pTRState->szPath, pTRState->szDefaultPage, MAX_URI );
				pTRState->szPath[MAX_URI - 1] = '\0';
			}
			// Combine the path with the base path.
			if( CombinePath( pTRState->szPath, MAX_URI, pTRState->szBasePath, CP_BEFORE ) == NULL ) {
				TRACE( _T( "Failed to combine with the base path.\n" ) );
				pTRState->szPath[0] = '\0';
				pTRState->state = trHead;
				return RQ_SERVER_ERROR;
			} else {
				TRACE( _T( "Combined with the base path: %s\n" ), (LPCTSTR)pTRState->szPath );
			}
			// Convert the path to full path.
			if( !ConvertToFullPath( pTRState->szFullPath, MAX_URI, pTRState->szPath ) ) {
				TRACE( _T( "Failed to convert to full path name.\n" ) );
				pTRState->szFullPath[0] = '\0';
				pTRState->state = trHead;
				return RQ_SERVER_ERROR;
			} else {
				TRACE( _T( "Converted to full path name: %s\n" ), (LPCTSTR)pTRState->szFullPath );
			}
			// Maintain these one-time variables.
			pTRState->fKeepAlive = false;
			pTRState->fIMS = false;
			pTRState->szIMS[0] = '\0';
			pTRState->szUserName[0] = '\0';
			pTRState->szPassword[0] = '\0';
			pTRState->getState = gtBeforeHeader;
			break;
		case gtBeforeHeader:
			// Ensure there is a complete line.
			if( (iResponseOne = TreatRequest_HasALine2( sRequest, piBegin, iEnd )) != RQ_SUCCESSFUL ) {
				return iResponseOne; // RQ_INCOMPLETE or RQ_ERROR
			}
			// Extract the header.
			if( (iResponseOne = ExtractWord( szBuffer, iBufferLength, sRequest, *piBegin, iEnd )) == 0 ) {
				// The line is empty.
				szBuffer[0] = '\0'; // Empty line.
			} else if( iResponseOne < 0 ) {
				// The word is too long.
				*piBegin -= iResponseOne;
				if( *piBegin >= SREQUESTBUFFER_SIZE ) {
					*piBegin -= SREQUESTBUFFER_SIZE;
				}
				TRACE( _T( "Format error (word too long)!\n" ) );
				pTRState->state = trHead;
				return RQ_ERROR;
			}
			*piBegin += iResponseOne;
			if( *piBegin >= SREQUESTBUFFER_SIZE ) {
				*piBegin -= SREQUESTBUFFER_SIZE;
			}
			if( strcmp( szBuffer, "Accept:" ) == 0 ) {
				pTRState->getState = gtAccept;
			} else if( strcmp( szBuffer, "Accept-Language:" ) == 0 ) {
				pTRState->getState = gtAcceptLanguage;
			} else if( strcmp( szBuffer, "Accept-Encoding:" ) == 0 ) {
				pTRState->getState = gtAcceptEncoding;
			} else if( strcmp( szBuffer, "User-Agent:" ) == 0 ) {
				pTRState->getState = gtUserAgent;
			} else if( strcmp( szBuffer, "Host:" ) == 0 ) {
				pTRState->getState = gtHost;
			} else if( strcmp( szBuffer, "Connection:" ) == 0 ) {
				pTRState->getState = gtConnection;
			} else if( strcmp( szBuffer, "If-Modified-Since:" ) == 0 ) {
				TRACE( _T( "Here, header If-Modified-Since received.\n" ) );
				pTRState->fIMS = true;
				pTRState->getState = gtIfModifiedSince;
			} else if( strcmp( szBuffer, "Authorization:" ) == 0 ) {
				pTRState->getState = gtAuthorization;
			} else if( szBuffer[0] == '\0' ) {
				// The line is an empty line.
				TRACE( _T( "End of request.\n" ) );
				goto labelAfterRequest;
			} else {
				// The header is unsupported. Ignore it.
				TRACE( _T( "Unsupported header: %s.\n" ), szBuffer );
				if( IgnoreLine( sRequest, piBegin, iEnd ) == RQ_INCOMPLETE ) {
					ASSERT( false );
					return RQ_INCOMPLETE;
				}
			}
			break;
		case gtAccept:
			// This header is not implemented. Ignore the line.
			if( iResponseOne = IgnoreLine( sRequest, piBegin, iEnd ) == RQ_INCOMPLETE ) {
				ASSERT( false );
				return RQ_INCOMPLETE;
			}
			pTRState->getState = gtBeforeHeader;
			break;
		case gtAcceptLanguage:
			if( iResponseOne = IgnoreLine( sRequest, piBegin, iEnd ) == RQ_INCOMPLETE ) {
				ASSERT( false );
				return RQ_INCOMPLETE;
			}
			pTRState->getState = gtBeforeHeader;
			break;
		case gtAcceptEncoding:
			if( iResponseOne = IgnoreLine( sRequest, piBegin, iEnd ) == RQ_INCOMPLETE ) {
				ASSERT( false );
				return RQ_INCOMPLETE;
			}
			pTRState->getState = gtBeforeHeader;
			break;
		case gtUserAgent:
			if( iResponseOne = IgnoreLine( sRequest, piBegin, iEnd ) == RQ_INCOMPLETE ) {
				ASSERT( false );
				return RQ_INCOMPLETE;
			}
			pTRState->getState = gtBeforeHeader;
			break;
		case gtHost:
			if( iResponseOne = IgnoreLine( sRequest, piBegin, iEnd ) == RQ_INCOMPLETE ) {
				ASSERT( false );
				return RQ_INCOMPLETE;
			}
			pTRState->getState = gtBeforeHeader;
			break;
		case gtConnection:
			// Test the connection token.
			// If it is "Keep-Alive" or "keep-alive", deal with it.
			{
				char szConnToken[30] = ""; // Connection token
				if( (iResponseOne = ExtractWord( szConnToken, 30, sRequest, *piBegin, iEnd )) <= 0 ) {
					// The word doesn't exist or the word is too long.
					// Ignore current line.
					IgnoreLine( sRequest, piBegin, iEnd );
					pTRState->state = trHead;
					// Return an error message.
					return RQ_ERROR;
				}
				// Ignore rest information on the line.
				if( IgnoreLine( sRequest, piBegin, iEnd ) == RQ_INCOMPLETE ) {
					ASSERT( false );
					return RQ_INCOMPLETE;
				}
				TRACE( _T( "Here, the connection token is: %s\n" ), szConnToken );
				// Turn szConnToken into lower case for comparison.
				strlwr( szConnToken );
				if( strcmp( szConnToken, "keep-alive" ) == 0 ) {
					pTRState->fKeepAlive = true;
				}
			}
			pTRState->getState = gtBeforeHeader;
			break;
		case gtIfModifiedSince:
			// Get the time string within the If-Modified-Since header.
			{
				int iBytesExtracted = 0;
				DiscardSpace( sRequest, piBegin, iEnd );
				// Extract the line.
				// The CR-LF marks are not stored in the string, but the
				// whole line including the CR-LF marks are extracted.
				if( (iResponseOne = ExtractLine( pTRState->szIMS, SZGMT_SIZE, sRequest, *piBegin, iEnd, &iBytesExtracted )) == RQ_INCOMPLETE ) {
					// Impossible.
					ASSERT( false );
					return RQ_INCOMPLETE;
				} else if( iResponseOne == RQ_ERROR ) {
					// The line is too long.
					// Ignore the line and return an error message.
					IgnoreLine( sRequest, piBegin, iEnd );
					pTRState->state = trHead;
					return RQ_ERROR;
				}
				*piBegin += iBytesExtracted;
				if( *piBegin >= SREQUESTBUFFER_SIZE ) {
					*piBegin -= SREQUESTBUFFER_SIZE;
				}
				TRACE( _T( "Here, If-Modified-Since time received: %s\n" ), pTRState->szIMS );
			}
			pTRState->getState = gtBeforeHeader;
			break;
		case gtAuthorization:
			// Store the password and username into a string.
			{
				char szAuthChallenge[MAX_USERNAME + MAX_PASSWORD] = ""; // Authorization challenge
				char szDecoded[MAX_USERNAME + MAX_PASSWORD] = "";
				// Extract the first word and check whether it is "Basic".
				if( (iResponseOne = ExtractWord( szAuthChallenge, MAX_USERNAME + MAX_PASSWORD, sRequest, *piBegin, iEnd )) <= 0 ) {
					// The word doesn't exist or is too long.
					// Ignore current line and return an error message.
					IgnoreLine( sRequest, piBegin, iEnd );
					pTRState->state = trHead;
					return RQ_ERROR;
				}
				TRACE( _T( "Authorization challenge type: %s\n" ), szAuthChallenge );
				if( strcmp( szAuthChallenge, "Basic" ) != 0 ) {
					// It is an unsupported challenge type.
					TRACE( _T( "Authorization challenge type not supported.\n" ) );
					// Ignore the line and to the end of the request.
					if( IgnoreLine( sRequest, piBegin, iEnd ) == RQ_INCOMPLETE ) {
						ASSERT( false );
						return RQ_INCOMPLETE;
					}
					goto labelAfterAuthChallenge; // Ignore the request.
				}
				*piBegin += iResponseOne;
				if( *piBegin >= SREQUESTBUFFER_SIZE ) {
					*piBegin -= SREQUESTBUFFER_SIZE;
				}
				// Extract the username-password string.
				if( (iResponseOne = ExtractWord( szAuthChallenge, MAX_USERNAME + MAX_PASSWORD, sRequest, *piBegin, iEnd )) <= 0 ) {
					// The word doesn't exist or is too long.
					TRACE( _T( "The authorization challenge entity cannot be retrieved.\n" ) );
					// Ignore this line and go to the end of the request.
					if( IgnoreLine( sRequest, piBegin, iEnd ) == RQ_INCOMPLETE ) {
						return RQ_INCOMPLETE;
					}
					goto labelAfterAuthChallenge;
				}
				*piBegin += iResponseOne;
				if( *piBegin >= SREQUESTBUFFER_SIZE ) {
					*piBegin -= SREQUESTBUFFER_SIZE;
				}
				TRACE( _T( "The authorization entity is: %s\n" ), szAuthChallenge );
				// Successfully extracted the authorization entity.
				// Ignore rest information on the same line.
				IgnoreLine( sRequest, piBegin, iEnd );
				// Decode the authorization entity.
				if( !CBase64::Decode( (BYTE *)szDecoded, MAX_USERNAME + MAX_PASSWORD, (BYTE *)szAuthChallenge ) ) {
					TRACE( _T( "Error decoding the authorization entity.\n" ) );
					goto labelAfterAuthChallenge;
				}
				TRACE(
					_T( "Successfully decoded the authorization entity.\n"
					"The result is: %s\n" ),
					szDecoded
				);
				// Extract the user name and the password from the decoded string.
				if( !ExtractUserPass( pTRState->szUserName, MAX_USERNAME,
					pTRState->szPassword, MAX_PASSWORD, szDecoded ) ) {
					// Failed to extract the strings.
					TRACE( _T( "Error extracting the user name and the password.\n" ) );
					// Clear the strings.
					pTRState->szUserName[0] = '\0';
					pTRState->szPassword[0] = '\0';
					goto labelAfterAuthChallenge;
				}
				TRACE( _T( "User name: %s\n" ), pTRState->szUserName );
				TRACE( _T( "Password: %s\n" ), pTRState->szPassword );
			}
labelAfterAuthChallenge:
			pTRState->getState = gtBeforeHeader;
			break;
		default:
			ASSERT( false );
			break;
		}
	}
labelAfterRequest:
	// Respond to/reply the request.
	//
	// The model of the following treatment:
	// Do One Step (e.g. open a file)
	// If error goto the corresponding treatment position
	// Do One Step
	// If error goto the corresponding treatment position
	// ...
	// ...
	// goto labelEnd
	// labelTreatmentOne:
	// ...
	// goto labelEnd
	// labelTreatmentTwo:
	// ...
	// goto labelEnd
	// ...
	// ...
	// labelEnd:
	// Do some cleanup or reset
	//
	{
		char szGMTFile[SZGMT_SIZE] = ""; // File time in GMT format
		char szGMTCurrent[SZGMT_SIZE] = ""; // Current time in GMT format
		char szResponse[600] = ""; // The response
		char szContentType[SZCONTENTTYPE_SIZE] = ""; // The content type
		char szExtension[SZEXTENSION_SIZE] = ""; // The extension of the file
		HANDLE hFileSend = INVALID_HANDLE_VALUE; // Handle to the file
		BY_HANDLE_FILE_INFORMATION fiSend; // File information
		SYSTEMTIME systimeFile; // File time in SYSTEMTIME format
		SYSTEMTIME systimeCurrent; // Current time in SYSTIME format
		DWORD iBytesRead = 0; // Number of file bytes read
		char szGMTIMS[SZGMT_SIZE] = ""; // RFC 1123 GMT of If-Modified-Since

		memset( &fiSend, 0, sizeof( fiSend ) );
		memset( &systimeFile, 0, sizeof( systimeFile ) );
		memset( &systimeCurrent, 0, sizeof( systimeCurrent ) );
		// Get current time and convert into GMT.
		// NOTE: SYSTEMTIME is already a GMT, but it is not a string.
		GetSystemTime( &systimeCurrent );
		SystemTimeToStringGMT( szGMTCurrent, &systimeCurrent );
		if( !IsValidFile( pTRState->szFullPath, pTRState->szBasePath ) ) {
			TRACE( _T( "The file path is invalid.\n" ) );
			goto labelErrorFile;
		}
		// Try to authorize the file.
		if( !Authorize( pTRState->szFullPath, pTRState->szBasePath, pTRState->szUserName, 
			pTRState->szPassword, pTRState->sarrPasswords, 
			pTRState->iPasswordCount, pTRState->plistMonitor,
			pTRState->pfRecToMon,
			pTRState->hwndDialog) ) 
		{
			TRACE( _T( "The file is unauthorized.\n" ) );
			goto labelUnauthorized;
		}
		// Open the file.
		hFileSend = CreateFile( pTRState->szFullPath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
		if( hFileSend == INVALID_HANDLE_VALUE ) {
			TRACE( _T( "Error opening file.\n" ) );
			goto labelErrorFile;
		}
		{
			CString sMonInfo;
			sMonInfo.Format(_T("File %s is being sent."), pTRState->szFullPath);
			SendMessage(pTRState->hwndDialog, MYWM_MONITOR_ADD_STRING, (WPARAM)0, (LPARAM)(LPCTSTR)sMonInfo);
		}
		// Extract the extension of the file.
		if( !ExtractExtension( szExtension, SZEXTENSION_SIZE, pTRState->szFullPath ) ) {
			TRACE( _T( "Error extracting file extension.\n" ) );
			CloseHandle( hFileSend );
			goto labelServerError;
		}
		TRACE( _T( "Extension retrieved: %s\n" ), szExtension );
		// Convert the extension to the content type.
		if( !ExtensionToContentType( szContentType, SZCONTENTTYPE_SIZE, szExtension ) ) {
			TRACE( _T( "Error converting file extension to content type.\n" ) );
			CloseHandle( hFileSend );
			goto labelServerError;
		}
		TRACE( _T( "Here, the content type is: %s\n" ), szContentType );
		// Get file information (mainly the time).
		if( !GetFileInformationByHandle( hFileSend, &fiSend ) ) {
			TRACE( _T( "Error getting file information.\n" ) );
			CloseHandle( hFileSend );
			goto labelServerError;
		}
		// Turn the file time into SYSTEMTIME format.
		FileTimeToSystemTime( &fiSend.ftLastWriteTime, &systimeFile );
		SystemTimeToStringGMT( szGMTFile, &systimeFile );
		// Compare to see whether to send a Not Modified message.
		if( pTRState->fIMS ) {
			// szIMS may not be in GMT format.
			// Convert szIMS into GMT format.
			HTTPDateTo1123(szGMTIMS, pTRState->szIMS);
			if (strcmp(szGMTIMS, szGMTFile) == 0) {
				TRACE( _T( "GMT compare match. I won't send the file.\n" ) );
				SendHTTPError( 304, pTRState->socketClient );
				goto labelAfterSendFile;
			} else {
				TRACE( _T( "GMT compare not match. I will still send the file.\n" ) );
			}
		}
		// Check whether the file is too large. This program is designed for
		// personal computers, so very large files are not supported.
		if( fiSend.nFileSizeHigh ) {
			TRACE( _T( "File too large.\n" ) );
			CloseHandle( hFileSend );
			goto labelServerError;
		}
		TRACE( _T( "GMT Generated: %s\n" ), szGMTCurrent );
		// Prepare and send the response.
		wsprintf( szResponse,
			_T(
				"HTTP/1.1 200 OK\r\n"
				"Server: Robbie Mosaic's HTTP Server\r\n"
				"Date: %s\r\n"
				"Content-Type: %s\r\n"
				"Accept-Ranges: bytes\r\n"
				"Last-Modified: %s\r\n"
				"Content-Length: %d\r\n"
				"\r\n"
			),
			szGMTCurrent,
			szContentType,
			szGMTFile,
			fiSend.nFileSizeLow
		);
		if( send( pTRState->socketClient, szResponse, strlen( szResponse ), 0 ) == SOCKET_ERROR ) {
			TRACE( _T( "Client socket error when sending response.\n" ) );
			goto labelAfterSendFile;
		}
		// If the request is HEAD but not GET, the file won't be sent.
		if( pTRState->fRequestHead ) goto labelAfterSendFile;
		// Send the file.
		do {
			if( ReadFile( hFileSend, pTRState->sSendBuffer, SENDBUFFER_SIZE, &iBytesRead, NULL ) ) {
				if( send( pTRState->socketClient, pTRState->sSendBuffer, iBytesRead, 0 ) == SOCKET_ERROR ) {
					TRACE( _T( "Client socket error when sending file.\n" ) );
					CloseHandle( hFileSend );
					goto labelAfterResponse;
				}
			}
		} while( iBytesRead == SENDBUFFER_SIZE );
		TRACE( _T( "File send complete.\n" ) );
labelAfterSendFile:
		CloseHandle( hFileSend );
		TRACE( _T( "File is closed.\n" ) );
		goto labelAfterResponse;
labelErrorFile:
		if( !SendHTTPError( 404, pTRState->socketClient ) ) {
			TRACE( _T( "Error when sending HTTP 404 Error.\n" ) );
		}
		goto labelAfterResponse;
labelUnauthorized:
		if( !SendHTTPError( 401, pTRState->socketClient ) ) {
			TRACE( _T( "Error when sending HTTP 401 Error.\n" ) );
		}
		goto labelAfterResponse;
labelServerError:
		if( !SendHTTPError( 500, pTRState->socketClient ) ) {
			TRACE( _T( "Error when sending HTTP 500 Error.\n" ) );
		}
		goto labelAfterResponse;
	}
labelAfterResponse:
	// Check whether the client required to keep the connection alive.
	if( !pTRState->fKeepAlive ) {
		// Shut the client down.
		// When recv returns, shut the server down and then close the socket.
		ShutSocketDown( pTRState->socketClient );
	}
	pTRState->state = trHead;
	return RQ_SUCCESSFUL; // All errors are dealt within the function.
}

void CServerThread::ShowRequest(const char *sRequest, int iBegin, int iEnd)
{
	char szBuffer[SREQUESTBUFFER_SIZE]; // The size must be enough.
	int i;
	int j;
	ASSERT( sRequest != NULL );
	// i is a subscript for sRequest.
	for( i = iBegin, j = 0; i != iEnd; j++ ) {
		szBuffer[j] = sRequest[i];
		// Increase i and manually round i up.
		if( ++i == SREQUESTBUFFER_SIZE ) i = 0;
	}
	szBuffer[j] = '\0';
	TRACE( _T( "Request: \n%s\n" ), szBuffer );
}

int ExtractWord(char *szWord, int iSize, const char *sRequest, int iBegin, int iEnd)
// Returns the number of bytes extracted.
// Parameter szWord can be NULL to discard information.
{
	int i = 0;
	int j = 0;
	char ch = ' ';
	bool fInsideWord = false;
	ASSERT(
		iSize >= 0 &&
		sRequest != NULL &&
		0 <= iBegin && iBegin < SREQUESTBUFFER_SIZE &&
		0 <= iEnd && iEnd < SREQUESTBUFFER_SIZE
	);
	// Here in the loop:
	// i - subscript for the word buffer.
	// j - subscript for the request buffer.
	fInsideWord = false;
	for( i = 0, j = iBegin; i < iSize && j != iEnd; ) {
		ch = sRequest[j];
		// When fInsideWord is false, turn it to true as soon as
		// the character is not a space.
		if( !fInsideWord && ch != ' ' && ch != '\t' ) {
			fInsideWord = true;
		}
		if( fInsideWord ) {
			if( ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n' ) break;
			if( szWord != NULL ) szWord[i] = ch;
			// Manually increase i.
			i++;
		}
		if( ++j == SREQUESTBUFFER_SIZE ) j = 0;
	}
	if( i == iSize ) {
		return -i; // Word too long.
	}
	if( j == iEnd ) {
		return 0; // Word incomplete. A word should end up with
			// a space or return (there is no other mark to
			// decide the end of the request).
	}
	if( szWord != NULL ) szWord[i] = '\0';
	if( j < iBegin ) {
		j += SREQUESTBUFFER_SIZE;
	}
	return j - iBegin; // Number of bytes extracted
}

// This function is obsolete and is not used.
int OldExtractLine(char *szLine, int iSize, const char *sRequest, int iBegin, int iEnd)
// Returns the length of the line.
// Parameter szLine can be NULL to discard information.
{
	int i = 0;
	int j = 0;
	char ch = ' ';
	bool fInsideLine = false;
	// Here in the loop:
	// i - subscript for the line buffer.
	// j - subscript for the request buffer.
	ASSERT( sRequest != NULL );
	for( i = 0, j = iBegin; i < iSize && j != iEnd; ) {
		ch = sRequest[j];
		if( !fInsideLine && ch != '\r' && ch != '\n' ) {
			fInsideLine = true;
		}
		if( fInsideLine ) {
			if( ch == '\r' || ch == '\n' ) break;
			if( szLine != NULL ) szLine[i] = ch;
			i++;
		}
		if( ++j == SREQUESTBUFFER_SIZE ) j = 0;
	}
	if( i == iSize ) {
		return -i; // Line too long.
	}
	if( j == iEnd ) {
		return 0; // Line incomplete. A line should end up with
			// a return (there is no other mark to
			// decide the end of the request).
	}
	if( szLine != NULL ) szLine[i] = '\0';
	if( j < iBegin ) {
		j += SREQUESTBUFFER_SIZE;
	}
	return j - iBegin;
}

int ExtractLine(char *szLine, int iSize, const char *sRequest, int iBegin, int iEnd, int *piBytesExtracted)
// Return values:
// RQ_SUCCESSFUL - A line is successfully extracted.
// RQ_INCOMPLETE - End of buffer reached. The extracted information may 
// not be complete.
// RQ_ERROR - Line too long or format error.
{
	int i = 0;
	int j = 0;
	char ch = ' ';
	// Possible values of pos:
	// 0 - The line of text.
	// 1 - At CR.
	// 2 - At LF. Break at this point.
	// 3 - Unexpected character.
	int pos = 0;
	ASSERT(
		iSize >= 0 &&
		piBytesExtracted != NULL &&
		0 <= iBegin && iBegin < SREQUESTBUFFER_SIZE &&
		0 <= iEnd && iEnd <= SREQUESTBUFFER_SIZE
	);
	
	for( pos = 0, i = 0, j = iBegin; i < iSize && j != iEnd; ) {
		ch = sRequest[j];
		switch( ch ) {
		case '\r':
			switch( pos ) {
			case 0: // Correct
				pos = 1;
				break;
			case 1: // Incorrect
				pos = 3;
				goto labelAfterCopyOne;
				break;
			}
			break;
		case '\n':
			switch( pos ) {
			case 0: // Incorrect
				pos = 2; // Ignore the error
				if( ++j == SREQUESTBUFFER_SIZE ) j -= SREQUESTBUFFER_SIZE;
				goto labelAfterCopyOne;
				break;
			case 1: // Correct
				pos = 2;
				if( ++j == SREQUESTBUFFER_SIZE ) j -= SREQUESTBUFFER_SIZE;
				goto labelAfterCopyOne;
				break;
			}
			break;
		default:
			switch( pos ) {
			case 0: // Correct
				pos = 0;
				break;
			case 1: // Incorrect
				pos = 3;
				goto labelAfterCopyOne;
				break;
			}
			break;
		}
		if( pos == 0 ) {
			if( szLine != NULL ) szLine[i] = sRequest[j];
			// Manually increase i.
			i++;
		}
		// Manually increase j and round j up.
		if( ++j == SREQUESTBUFFER_SIZE ) j = 0;
	}
labelAfterCopyOne:
	if( i == iSize ) {
		// Overflow
		*piBytesExtracted = iSize;
		return RQ_ERROR;
	}
	if( szLine != NULL ) szLine[i] = '\0';
	*piBytesExtracted = j - iBegin;
	if( *piBytesExtracted < 0 ) *piBytesExtracted += SREQUESTBUFFER_SIZE;
	if( pos < 2 ) {
		return RQ_INCOMPLETE; // End of buffer reached.
	} else if( pos == 2 ) {
		return RQ_SUCCESSFUL;
	} else { // pos == 3
		return RQ_ERROR;
	}
}

int DiscardSpace(const char *sRequest, int *piBegin, int iEnd)
{
	int i;
	char ch;
	for( i = *piBegin; i != iEnd; ) {
		ch = sRequest[i];
		if( ch != ' ' && ch != '\t' ) break;
		// Manually increase i and round i up.
		if( ++i == SREQUESTBUFFER_SIZE ) i = 0;
	}
	if( i == SREQUESTBUFFER_SIZE ) {
		i = 0;
	}
	*piBegin = i;
	return i;
}

int CServerThread::IgnoreLine(const char *sRequest, int *piBegin, int iEnd)
{
	int iCopyOfEnd = iEnd;
	int iBytesExtracted = 0;
	int iResponseOne = RQ_SUCCESSFUL;
	ASSERT( sRequest != NULL && piBegin != NULL && 0 <= iEnd && iEnd < SREQUESTBUFFER_SIZE );
	if( (iResponseOne = ExtractLine( NULL, SREQUESTBUFFER_SIZE + 1, sRequest, *piBegin, iEnd, &iBytesExtracted )) == RQ_INCOMPLETE ) {
		// Within the buffer the end of line is not seen.
		// Test whether the buffer is full.
		iCopyOfEnd++;
		if( iCopyOfEnd >= SREQUESTBUFFER_SIZE ) iCopyOfEnd -= SREQUESTBUFFER_SIZE;
		// If the buffer is full, iCopyOfEnd should now equal 
		// to *piBegin.
		if( iCopyOfEnd == *piBegin ) {
			// The buffer is full.
			// The whole buffer is scanned but the end of line
			// is not found.
			// It means the line is so long that it is illegal.
			*piBegin = iEnd; // Empty the buffer.
			return RQ_SUCCESSFUL;
		} else {
			return RQ_INCOMPLETE;
		}
	} else if( iResponseOne == RQ_SUCCESSFUL ) {
		// Successfully passed the line.
		*piBegin += iBytesExtracted;
		if( *piBegin >= SREQUESTBUFFER_SIZE ) *piBegin -= SREQUESTBUFFER_SIZE;
	}
#ifdef _DEBUG
	else if( iResponseOne == RQ_ERROR ) {
		// For debug only.
		// As SREQUESTBUFFER_SIZE + 1 is specified as the maximum size
		// for scanning, it should always be enough.
		// It should be impossible.
		ASSERT( false );
	}
#endif
	return RQ_SUCCESSFUL;
}

int OldExtractWord2(char *szWord, int iSize, const char *szLine, int iBegin)
{
	int i = 0;
	int j = 0;
	char ch = ' ';
	bool fInsideWord = false;
	ASSERT( szLine != NULL && iSize >= 0 && iBegin >= 0 );
	// Here in the loop:
	// i - subscript for the word buffer.
	// j - subscript for the request buffer.
	for( i = 0, j = iBegin; i < iSize && szLine[j]; j++ ) {
		ch = szLine[j];
		if( !fInsideWord && ch != ' ' && ch != '\t' ) {
			fInsideWord = true;
		}
		if( fInsideWord ) {
			if( ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n' ) break;
			if( szWord != NULL ) szWord[i] = ch;
			// Manually increase i.
			i++;
		}
	}
	if( i == iSize ) {
		return -i; // Word too long.
	}
	if( szWord != NULL ) szWord[i] = '\0';
	return j - iBegin;
}

int ExtractWord2(TCHAR *szWord, int iSize, const TCHAR *szLine, int iBegin)
// This function is compatible with MBCS.
{
	bool fInsideWord;
	const TCHAR *pchSrc = NULL;
	TCHAR *pchDst = NULL;
	bool fOverflow;
	ASSERT(szLine != NULL && iSize >= 0 && iBegin >= 0);
	// Here in the loop:
	// pchSrc - pointer for the source buffer.
	// pchDst - pointer for the destination buffer.
	pchSrc = szLine + iBegin;
	pchDst = szWord;
	fInsideWord = false;
	fOverflow = false;
	while (*pchSrc) {
		if (!fInsideWord && *pchSrc != _T(' ') && *pchSrc != _T('\t')) {
			fInsideWord = true;
		}
		if (fInsideWord) {
			if (*pchSrc == _T(' ') || *pchSrc == _T('\t') ||
				*pchSrc == _T('\r') || *pchSrc == _T('\n')) {
				break;
			}
			if (_tcsinc(pchSrc) - pchSrc == 2) {
				// The character to copy is a two-byte character.
				if (pchDst + 2 - szWord >= iSize) {
					// Over flow. "+ 2" means the zero after the two bytes reserved 
					// for current character.
					fOverflow = true;
					break;
				}
				// Check whether the pointer is NULL to allow "virtual write".
				if (szWord != NULL) {
					*pchDst = *pchSrc;
					*(pchDst + 1) = *(pchSrc + 1);
				}
				pchDst += 2;
			} else { // The character takes one byte.
				if (pchDst + 1 - szWord >= iSize) {
					fOverflow = true;
					break;
				}
				if (szWord != NULL) *pchDst = *pchSrc;
				pchDst++;
			}
		}
		pchSrc = _tcsinc(pchSrc);
	}
	if( fOverflow ) {
		return -iSize; // Word too long.
	}
	// Complete the destination string.
	if( szWord != NULL ) *pchDst = _T('\0');
	return pchSrc - (szLine + iBegin);
}

bool CServerThread::TreatRequest_HasALine(const char *sRequest, int iBegin, int iEnd)
{
	int i = 0;
	bool fContainFullLine = false;
	ASSERT( sRequest != NULL && 0 <= iBegin && iBegin < SREQUESTBUFFER_SIZE &&
		0 <= iEnd && iEnd < SREQUESTBUFFER_SIZE );
	// Check to see whether sRequest contains a full line.
	for( i = iBegin; i != iEnd; )
	{
		if( sRequest[i] == '\n' ) {
			fContainFullLine = true;
			break;
		}
		// Manually increase i and round i up.
		if( ++i == SREQUESTBUFFER_SIZE ) i = 0;
	}
	return fContainFullLine;
}

int CServerThread::ConvertURIToPath(char *szPath, int iSize, const char *szURI)
{
	int i = 0;
	// Decide the first character ('h' or '/')
	if( szURI[0] == 'h' ) {
		// Pass the following letters ("ttp").
		for( i = 0; szURI[i]; i++ ) {
			if( !isalpha( szURI[i] ) ) {
				break;
			}
		}
		// ':'
		if( szURI[i] != ':' ) {
			return RQ_ERROR;
		}
		i++;
		// '/'
		if( szURI[i] != '/' ) {
			return RQ_ERROR;
		}
		i++;
		// '/'
		if( szURI[i] != '/' ) {
			return RQ_ERROR;
		}
		i++;
		// Pass the host name ("www.abc.com").
		for( ; szURI[i]; i++ ) {
			if( szURI[i] == '/' ) {
				break;
			}
		}
		// Check wether the character is '/' (perhaps it's the end of the string).
		if( szURI[i] != '/' ) {
			return RQ_ERROR;
		}
	} else if( szURI[0] != '/' ) {
		// If the first character is not '/'
		return RQ_ERROR;
	}
	// Pass the first '/'.
	i++;

	int j = 0;
	char ch = ' ';
	char cHexHigh = ' ', cHexLow = ' '; // Characters as the buffer for %XX format character
	bool fInHexMode = false; // Whether in %XX format
	// Convert the URI to path.
	for( j = 0; szURI[i] && j < iSize; i++ ) {
		ch = szURI[i];
		if( !fInHexMode ) {
			switch( ch ) {
			case '%': // %XX
				fInHexMode = true;
				cHexHigh = ' ';
				cHexLow = ' ';
				break;
			case '/': // Convert to '\\'
				szPath[j] = '\\';
				// Manually increase j.
				j++;
				break;
			case '?': // A question mark ends the path part of a URI
				szPath[j] = '\0';
				j++;
				break;
			default: // Copy the character
				szPath[j] = ch;
				// Manually increase j.
				j++;
				break;
			}
		} else { // Already in %XX
			switch( ch ) {
			case '%': // The second '%' in "%%".
				// "%%" is converted to '%'.
				fInHexMode = false;
				szPath[j] = ch;
				// Manually increase j.
				j++;
				break;
			default:
				if( cHexHigh == ' ' ) { // The first HEX digit.
					if( !isxdigit( ch ) ) { // It should be a HEX digit.
						return RQ_ERROR;
					}
					cHexHigh = ch;
				} else { // The second HEX digit.
					if( !isxdigit( ch ) ) {
						return RQ_ERROR;
					}
					cHexLow = ch;
					// Convert the digits to a character.
					szPath[j] = hdtochar( cHexHigh, cHexLow );
					// Manually increase j.
					j++;
					fInHexMode = false;
				}
				break;
			}
		}
	}
	if( fInHexMode ) return RQ_ERROR; // Broken while interpreting digits
	if( j == iSize ) {
		// Not enough buffer
		return RQ_SERVER_ERROR;
	}
	szPath[j] = '\0';
	return RQ_SUCCESSFUL;
}

char hdtochar(char cHexDigitHigh, char cHexDigitLow)
{
	char ch;
	ASSERT( isxdigit( cHexDigitHigh ) && isxdigit( cHexDigitLow ) );
	if( isdigit( cHexDigitHigh ) ) {
		ch = (cHexDigitHigh - '0') << 4;
	} else {
		if( isupper( cHexDigitHigh ) ) {
			ch = (cHexDigitHigh - 'A' + 10) << 4;
		} else {
			ch = (cHexDigitHigh - 'a' + 10) << 4;
		}
	}
	if( isdigit( cHexDigitLow ) ) {
		ch |= cHexDigitLow - '0';
	} else {
		if( isupper( cHexDigitLow ) ) {
			ch |= cHexDigitLow - 'A' + 10;
		} else {
			ch |= cHexDigitLow - 'a' + 10;
		}
	}
	return ch;
}

// CombinePath
// Combines two paths, one is relative to the other.
// iMethod indicates whether to put the second after the first or
// to put the second before the first.
char *CombinePath(TCHAR *szPath, int iSize, const TCHAR *szPathOther, int iMethod)
// This function is compatible with MBCS.
{
	ASSERT( szPath != NULL && iSize >= 0 && szPathOther != NULL );
	switch( iMethod ) {
	case CP_AFTER:
		{
			int i = 0;
			int j = 0;
			TCHAR *pchOne = NULL;

			i = strlen( szPath );
			if( i > 0 ) {
				// The length is above zero. Find the last character.
				pchOne = &szPath[i]; // The null character.
				pchOne = _tcsdec(szPath, pchOne); // The last character.
				if (*pchOne != _T('\\')) {
					// The last character is not '\\' so try to fix a '\\' after the path.
					if( i + 1 >= iSize ) { // Buffer too small
						return NULL;
					}
					szPath[i] = _T('\\');
					szPath[i + 1] = _T('\0');
					i++; // Point to the null character.
				}
			} else { // i == 0
				// Try to fix a '\\' after the path.
				if( i + 1 >= iSize ) {
					return NULL;
				}
				szPath[i] = _T('\\');
				szPath[i + 1] = _T('\0');
				i++;
			}
			j = 0;
			// Skip the first '\\' if it exists.
			if (szPathOther[j] == _T('\\')) j++;
			// Go on treating. Copy the string.
			for( ; i < iSize && szPathOther[j]; i++, j++ ) {
				szPath[i] = szPathOther[j];
			}
			if( i == iSize ) { // Buffer too small
				return NULL;
			}
			// Complete the string.
			szPath[i] = _T('\0');
			return szPath;
		}
		break;
	case CP_BEFORE:
		{
			char szBuffer[MAX_URI]; // Use szBuffer to temporarily hold the result
			int i = 0;
			int j = 0;
			TCHAR *pchOne = NULL;

			// Copy information from szPathOther to szBuffer.
			for( i = 0, j = 0; i < MAX_URI && szPathOther[j]; i++, j++ ) {
				szBuffer[i] = szPathOther[j];
			}
			if( i == MAX_URI ) { // Buffer szBuffer is too small.
				return NULL;
			}
			szBuffer[i] = _T('\0');
			if( i > 0 ) {
				pchOne = &szBuffer[i];
				pchOne = _tcsdec(szBuffer, pchOne);
				if (*pchOne != _T('\\')) {
					if( i + 1 >= MAX_URI ) { // Buffer szBuffer is too small.
						return NULL;
					}
					szBuffer[i] = _T('\\');
					szBuffer[i + 1] = _T('\0');
					i++;
				}
			} else { // i == 0
				if( i + 1 >= MAX_URI ) {
					return NULL;
				}
				szBuffer[i] = _T('\\');
				szBuffer[i + 1] = _T('\0');
				i++;
			}
			j = 0;
			if (szPath[j] == _T('\\')) j++;
			for( ; i < MAX_URI && szPath[j]; i++, j++ ) {
				szBuffer[i] = szPath[j];
			}
			if( i == MAX_URI ) { // Buffer szBuffer is too small.
				return NULL;
			}
			szBuffer[i] = _T('\0');
			for( i = 0; i < iSize && szBuffer[i]; i++ ) {
				szPath[i] = szBuffer[i];
			}
			if( i == iSize ) { // Buffer too small
				return NULL;
			}
			szPath[i] = _T('\0');
			return szPath;
		}
		break;
	default:
		ASSERT( false );
		return NULL;
		break;
	}
}

char *SystemTimeToStringGMT(char *szGMT, const SYSTEMTIME *psystimeOne)
{
	const char szarrWeekDays[][4] = {
		"Sun",
		"Mon",
		"Tue",
		"Wed",
		"Thu",
		"Fri",
		"Sat"
	};
	const char szarrMonths[][4] = {
		"Jan",
		"Feb",
		"Mar",
		"Apr",
		"May",
		"Jun",
		"Jul",
		"Aug",
		"Sep",
		"Oct",
		"Nov",
		"Dec"
	};
	ASSERT( szGMT != NULL && psystimeOne != NULL );

	wsprintf( szGMT, "%s, %02hu %s %04hu %02hu:%02hu:%02hu GMT",
		szarrWeekDays[psystimeOne->wDayOfWeek],
		psystimeOne->wDay,
		szarrMonths[psystimeOne->wMonth - 1],
		psystimeOne->wYear,
		psystimeOne->wHour,
		psystimeOne->wMinute,
		psystimeOne->wSecond
	);
	return szGMT;
}

bool ExtractExtension(TCHAR *szExtension, int iSize, const TCHAR *szPath)
// This function is compatible with MBCS.
{
	const TCHAR *pchSrc = NULL;
	TCHAR *pchDst = NULL;
	bool fExtension = false;
	ASSERT( szExtension != NULL && iSize >= 0 );

	// Find the rightmost '\\' or the rightmost '.'.
	if (!szPath[0]) {
		// The string is empty.
		if (iSize < 1) {
			// The size is not large enough.
			return false;
		}
		szExtension[0] = _T('\0');
		return true;
	}
	pchSrc = _tcsdec(szPath, szPath + strlen(szPath)); // Point to the last character.
	while (true) {
		if (*pchSrc == _T('\\')) {
			fExtension = false;
			break;
		} else if (*pchSrc == _T('.')) {
			fExtension = true;
			break;
		}
		if (pchSrc == szPath) {
			fExtension = false;
			break; // The first character has been treated
		}
		pchSrc = _tcsdec(szPath, pchSrc);
	}
	if (!fExtension) { // No extension
		szExtension[0] = _T('\0');
		return true;
	}
	// There is an extension. Make pchSrc point to the character next to '.'.
	bool fOverflow = false;
	pchSrc = _tcsinc(pchSrc);
	pchDst = szExtension;
	fOverflow = false;
	while (*pchSrc) {
		if (_tcsinc(pchSrc) - pchSrc == 2) {
			// The character to copy is a two-byte character.
			if (pchDst + 2 - szExtension >= iSize) {
				// Overflow. "+ 2" means the null character after the two bytes reserved 
				// for current character.
				fOverflow = true;
				break;
			}
			*pchDst = *pchSrc;
			*(pchDst + 1) = *(pchSrc + 1);
			pchDst += 2;
		} else { // The character takes one byte.
			if (pchDst + 1 - szExtension >= iSize) {
				fOverflow = true;
				break;
			}
			*pchDst = *pchSrc;
			pchDst++;
		}
		pchSrc = _tcsinc(pchSrc);
	}
	if (fOverflow) {
		return false;
	}
	*pchDst = _T('\0');
	return true;
}

bool ExtensionToContentType(char *szContentType, int iSize, const char *szExtension)
{
	HKEY hKeyExtension; // Key of the specified extension
	char szRegKey[128] = ""; // Path of the registry key
	int i = 0;
	int j = 0;
	DWORD dwType;
	DWORD iBytes = iSize;
	char szBuffer[SZEXTENSION_SIZE];
	ASSERT( szContentType != NULL && iSize >= 0 && szExtension != NULL );

	// Make the name of the registry key ".ext".
	wsprintf( szRegKey, _T(".%s"), szExtension );
	// Open the registry key.
	if( RegOpenKey( HKEY_CLASSES_ROOT, szRegKey, &hKeyExtension ) != ERROR_SUCCESS ) {
		// The operation failed.
		TRACE( _T( "Error opening registry key: HKEY_CLASSES_ROOT\\%s\n" ), szRegKey );
		return false;
	}
	// Get the string value whose name is "Content Type".
	if( RegQueryValueEx( hKeyExtension, "Content Type", NULL, &dwType, (LPBYTE)szContentType, &iBytes ) != ERROR_SUCCESS ) {
		// The value doesn't exist.
		TRACE( _T( "Error getting registry data.\n" ) );
		// Close the key.
		RegCloseKey( hKeyExtension );
		// Use the default values instead.
		TRACE( _T( "Use automatic default values for the content type.\n" ) );
		strncpy( szBuffer, szExtension, SZEXTENSION_SIZE );
		szBuffer[SZEXTENSION_SIZE - 1] = '\0';
		strlwr( szBuffer );
		if( strcmp( szBuffer, "htm" ) == 0 || strcmp( szBuffer, "html" ) == 0 ) {
			strncpy( szContentType, "text/html", iSize );
			szContentType[iSize - 1] = '\0';
		} else if( strcmp( szBuffer, "vbs" ) == 0 ) {
			strncpy( szContentType, "text/plain", iSize );
			szContentType[iSize - 1] = '\0';
		} else if( strcmp( szBuffer, "js" ) == 0 ) {
			strncpy( szContentType, "text/plain", iSize );
			szContentType[iSize - 1] = '\0';
		} else if( strcmp( szBuffer, "css" ) == 0 ) {
			strncpy( szContentType, "text/css", iSize );
			szContentType[iSize - 1] = '\0';
		} else if( strcmp( szBuffer, "mid" ) == 0 || strcmp( szBuffer, "rmi" ) == 0 ) {
			strcmp( szContentType, "audio/mid" );
		} else {
			strncpy( szContentType, "application/x-msdownload", iSize );
			szContentType[iSize - 1] = '\0';
		}
		return true;
	}
	RegCloseKey( hKeyExtension );
	// Successfully got the registry data.
	// Check the type.
	if( dwType != REG_SZ && dwType != REG_EXPAND_SZ ) {
		TRACE( _T( "The type of the registry key is not right.\n" ) );
		return false;
	}
	// Check the size.
	if( (int)iBytes == iSize ) {
		TRACE( _T( "The size of the registry key is too large.\n" ) );
		return false;
	}
	// Successful.
	szContentType[iBytes] = '\0';
	return true;
}

bool CServerThread::IsValidFile(const char *szFullPath, const char *szBasePath)
{
	// This function doesn't recurse or call recursive functions, so let
	// the following strings be in the stack.
	char szBufferOne[MAX_URI];
	char szBufferTwo[MAX_URI];
	bool fResult = false;
	int i = 0;
	ASSERT( szFullPath != NULL && szBasePath != NULL );
	// Under Windows, file names should be compared in the same case.
	// First copy the path names into the buffers.
	strncpy( szBufferOne, szFullPath, MAX_URI );
	szBufferOne[MAX_URI - 1] = '\0';
	strncpy( szBufferTwo, szBasePath, MAX_URI );
	szBufferTwo[MAX_URI - 1] = '\0';
	// Then make both strings into lower case.
	strlwr( szBufferOne );
	strlwr( szBufferTwo );
	// Compare the two strings.
	fResult = true;
	for( i = 0; szBufferTwo[i] && szBufferOne[i]; i++ ) {
		if( szBufferTwo[i] != szBufferOne[i] ) {
			fResult = false; // They are different so the file path
				// is of course invalid.
			break;
		}
	}
	if( fResult && szBufferTwo[i] != '\0' ) {
		// The base path is longer than the file path so the file
		// path is also invalid.
		fResult = false;
	}
	return fResult;
}

bool CServerThread::SendHTTPError(int iError, SOCKET socketClient)
{
	char szResponse[600] = "";
	const char szNotModifiedHead[] = "304 Not Modified";
	const char szNotFound[] = "<html><body><h2>HTTP 404: Not Found</h2></body></html>\r\n";
	const char szNotFoundHead[] = "404 Not Found";
	const char szBadRequest[] = "<html><body<h2>HTTP 400: Bad Request</h2></body></html>\r\n";
	const char szBadRequestHead[] = "400 Bad Request";
	const char szUnauthorized[] = "<html><body><h2>HTTP 401: Unauthorized</h2></body></html>\r\n";
	const char szUnauthorizedHead[] = "401 Unauthorized";
	const char szNotImplemented[] = "<html><body><h2>HTTP 501: Not Implemented</h2></body></html>\r\n";
	const char szNotImplementedHead[] = "501 Not Implemented";
	const char szInternalServerError[] = "<html><body><h2>HTTP 500: Internal Server Error</h2></body></html>\r\n";
	const char szInternalServerErrorHead[] = "500 Internal Server Error";
	char szGMTCurrent[SZGMT_SIZE] = "";
	SYSTEMTIME systimeCurrent;

	// Current time is needed in the information to return.
	GetSystemTime( &systimeCurrent );
	SystemTimeToStringGMT( szGMTCurrent, &systimeCurrent );
	switch( iError ) {
	case 304: // Not Modified
		wsprintf( szResponse,
			_T(
				"HTTP/1.1 %s\r\n"
				"Server: Robbie Mosaic's HTTP Server\r\n"
				"Date: %s\r\n"
				"Content-Type: text/html\r\n"
				"Content-Length: 0\r\n"
				"\r\n"
			),
			szNotModifiedHead,
			szGMTCurrent
		);
		break;
	case 400:
		TRACE( _T( "Bad request.\n" ) );
		wsprintf( szResponse,
			_T(
				"HTTP/1.1 %s\r\n"
				"Server: Robbie Mosaic's HTTP Server\r\n"
				"Date: %s\r\n"
				"Content-Type: text/html\r\n"
				"Content-Length: %d\r\n"
				"\r\n"
				"%s"
			),
			szBadRequestHead,
			szGMTCurrent,
			strlen(szBadRequest),
			szBadRequest
		);
		break;
	case 401:
		wsprintf( szResponse,
			_T(
				"HTTP/1.1 %s\r\n"
				"Server: Robbie Mosaic's HTTP Server\r\n"
				"Date: %s\r\n"
				"Content-Type: text/html\r\n"
				"Content-Length: %d\r\n"
				"WWW-Authenticate: Basic\r\n"
				"\r\n"
				"%s"
			),
			szUnauthorizedHead,
			szGMTCurrent,
			strlen(szUnauthorized),
			szUnauthorized
		);
		break;
	case 404:
		wsprintf( szResponse,
			_T(
				"HTTP/1.1 %s\r\n"
				"Server: Robbie Mosaic's HTTP Server\r\n"
				"Date: %s\r\n"
				"Content-Type: text/html\r\n"
				"Content-Length: %d\r\n"
				"\r\n"
				"%s"
			),
			szNotFoundHead,
			szGMTCurrent,
			strlen(szNotFound),
			szNotFound
		);
		break;
	case 500:
		wsprintf( szResponse,
			_T(

				"HTTP/1.1 %s\r\n"
				"Server: Robbie Mosaic's HTTP Server\r\n"
				"Date: %s\r\n"
				"Content-Type: text/html\r\n"
				"Content-Length: %d\r\n"
				"\r\n"
				"%s"
			),
			szInternalServerErrorHead,
			szGMTCurrent,
			strlen(szInternalServerError),
			szInternalServerError
		);
		break;
	case 501:
		wsprintf( szResponse,
			_T(

				"HTTP/1.1 %s\r\n"
				"Server: Robbie Mosaic's HTTP Server\r\n"
				"Date: %s\r\n"
				"Content-Type: text/html\r\n"
				"Content-Length: %d\r\n"
				"\r\n"
				"%s"
			),
			szNotImplementedHead,
			szGMTCurrent,
			strlen(szNotImplemented),
			szNotImplemented
		);
		break;
	default:
		ASSERT( false );
		break;
	}
	if( send( socketClient, szResponse, strlen( szResponse ), 0 ) == SOCKET_ERROR ) {
		return false;
	}
	return true;
}

bool ConvertToFullPath(char *szFullPath, int iSize, const char *szPath)
{
	char *pcFNComponent = NULL;
	ASSERT( szFullPath != NULL && iSize >= 0 && szPath != NULL );
	if( !GetFullPathName( szPath, iSize, szFullPath, &pcFNComponent ) ) {
		return false;
	}
	return true;
}

bool CServerThread::Authorize(
							  const char *szFullPath,
							  const char *szBasePath,
							  const char *szUserName,
							  const char *szPassword,
							  CString *sarrPasswords,
							  int iPasswordCount,
							  CListBox *plistMonitor,
							  BOOL *pfRecToMon,
							  HWND hwndDialog
							  )
{
	int i = 0;
	int j = 0;
	int iLenFullPath = strlen( szFullPath );
	// This function doesn't recurse or call recursive functions.
	// Let the following variables be in the stack.
	char szPartialPath[MAX_URI];
	char szPartialPath2[MAX_URI];
	char szPartialPath3[MAX_URI];
	bool fReturnValue = true;
	char szUserName2[MAX_USERNAME] = "";
	char szPassword2[MAX_PASSWORD] = "";
	int iPasswordNumber = 0;
	int iWordStart = 0;
	bool fAuthorizationRequired = false;
	ASSERT( szFullPath != NULL && szBasePath != NULL &&
		szUserName != NULL && szPassword != NULL &&
		((sarrPasswords != NULL && iPasswordCount > 0) ||
		(sarrPasswords == NULL && iPasswordCount == 0)) &&
		plistMonitor != NULL );

	szPartialPath[0] = '\0';
	szPartialPath2[0] = '\0';
	szPartialPath3[0] = '\0';
	i = strlen( szBasePath );
	if( i == 0 ) {
		TRACE( _T( "Authorization error: the length of szBasePath is zero.\n" ) );
		fReturnValue = false;
		goto labelBeforeReturn;
	}
	// First copy the partial path (not including '\\').
	if( szBasePath[i - 1] != '\\' ) {
		i++;
	}
	for( j = 0; i < iLenFullPath && j < MAX_URI; i++, j++ ) {
		szPartialPath[j] = szFullPath[i];
	}
	if( j >= MAX_URI ) {
		TRACE( _T( "Authorization error: szFullPath is too long.\n" ) );
		fReturnValue = false;
		goto labelBeforeReturn;
	}
	szPartialPath[j] = '\0';
	// For each password list entry:
	for( iPasswordNumber = 0; iPasswordNumber < iPasswordCount; iPasswordNumber++ ) {
		// Extract the partial path (in the form of URI).
		if( (iWordStart = ExtractWord2( szPartialPath2, MAX_URI, (LPCTSTR)sarrPasswords[iPasswordNumber], 0 )) <= 0 ) {
			// Failed to extract partial path.
			szPartialPath2[0] = '\0';
			TRACE( _T( "Authorization error when extracting partical paths from the password list.\n" ) );
		}
		// Convert it into path form.
		if( ConvertURIToPath( szPartialPath3, MAX_URI, szPartialPath2 ) != RQ_SUCCESSFUL ) {
			// Failed to convert URI to path.
			szPartialPath3[0] = '\0';
			TRACE( _T( "Authorization error converting URI to path when authorizing.\n" ) );
		}
		// Compare both partial paths in lower case.
		if( stricmp( szPartialPath, szPartialPath3 ) == 0 ) {
			// Check whether the user names and passwords match.
			// iWordStart has been initialized with the first word.
			if( (i = ExtractWord2( szUserName2, MAX_USERNAME, (LPCTSTR)sarrPasswords[iPasswordNumber], iWordStart )) <= 0 ) {
				// Error extracting the user name.
				TRACE( _T( "Authorization error: failed to extract the second word.\n" ) );
				fReturnValue = true;
				goto labelBeforeReturn;
			}
			// Increase iWordStart.
			iWordStart += i;
			// Compare the user names in lower case.
			if( stricmp( szUserName2, szUserName ) != 0 ) {
				// The user names are different.
				// Try to find another entry for the same file.
				fAuthorizationRequired = true;
				goto labelAfterOnceCheck;
			}
			// Extract the password.
			if( (i = ExtractWord2( szPassword2, MAX_PASSWORD, (LPCTSTR)sarrPasswords[iPasswordNumber], iWordStart )) <= 0 ) {
				TRACE( _T( "Authorization error: failed to extract the third word.\n" ) );
				// Failed to extract the password.
				// For safety for the file with a user name, don't allow access.
				fReturnValue = false;
				goto labelBeforeReturn;
			}
			// Compare both passwords.
			if( strcmp( szPassword, szPassword2 ) != 0 ) {
				// Password not match. The access is denied.
				fReturnValue = false;
				goto labelBeforeReturn;
			}
			// Authorization successfully passed.
			fReturnValue = true;
			goto labelBeforeReturn;
		}
labelAfterOnceCheck:
		;
	}
	// None of the checks matched.
	// If once the file name is matched, it is required to be authorized.
	if( fAuthorizationRequired ) {
		TRACE( _T( "The user name is unknown:%s\n" ), szUserName );
		fReturnValue = false;
	} else {
		TRACE( _T( "The file doesn't need authorization:\n%s\n" ), szPartialPath );
		fReturnValue = true;
	}
labelBeforeReturn:
	if( fReturnValue ) {
		// Send monitor information
		{
			CString sMonitor;
			sMonitor.Format( _T( "The file: "
				"%s "
				"is authorized." ),
				szFullPath
			);
			SendMessage(hwndDialog, MYWM_MONITOR_ADD_STRING, (WPARAM)0, (LPARAM)(LPCTSTR)sMonitor);
		}
	} else {
		{
			CString sMonitor;
			sMonitor.Format( _T( "The file: "
				"%s "
				"is unauthorized. Access to the file is denied." ),
				szFullPath
			);
			SendMessage(hwndDialog, MYWM_MONITOR_ADD_STRING, (WPARAM)0, (LPARAM)(LPCTSTR)sMonitor);
		}
	}
	return fReturnValue;
}

bool CServerThread::ExtractUserPass(char *szUserName, int iSizeUser, char *szPassword, int iSizePass, const char *szUserPass)
{
	int i;
	int j;
	ASSERT( szUserName != NULL && iSizeUser >= 0 && szPassword != NULL && iSizePass >= 0 && szUserPass != NULL );
	for( i = 0; szUserPass[i] && szUserPass[i] != ':' && i < iSizeUser; i++ ) {
		szUserName[i] = szUserPass[i];
	}
	if( i >= iSizeUser ) {
		return false;
	}
	szUserName[i] = '\0';
	if( szUserPass[i] == '\0' ) {
		// It should be ':', the separator, but not '\0'.
		return false;
	}
	i++;
	for( j = 0; szUserPass[i] && j < iSizePass; i++, j++ ) {
		szPassword[j] = szUserPass[i];
	}
	if( j >= iSizePass ) {
		return false;
	}
	szPassword[j] = '\0';
	return true;
}

bool CServerThread::ShutSocketDown(SOCKET socketOne)
{
	bool fSuccessful = true;
	if( shutdown( socketOne, SD_SEND ) == SOCKET_ERROR ) {
		TRACE( _T( "Socket error when shutting down.\n" ) );
		fSuccessful = false;
	}
	return fSuccessful;
}

int CServerThread::TreatRequest_HasALine2(const char *sRequest, int *piBegin, int iEnd)
{
	if( !TreatRequest_HasALine( sRequest, *piBegin, iEnd ) ) 
	{
		// Check whether the buffer is full.
		if( (*piBegin == 0 && iEnd == SREQUESTBUFFER_SIZE - 1) ||
			(*piBegin != 0 && iEnd == *piBegin - 1) ) {
			// The buffer is full. Return an error message.
			*piBegin = iEnd;
			return RQ_ERROR;
		} else {
			// The buffer is not full.
			return RQ_INCOMPLETE;
		}
	}
	return RQ_SUCCESSFUL;
}

bool HTTPDateTo1123(char *szRFC1123, const char *szHTTPDate)
{
	char szWord[16];
	int i;
	int noc; // number of characters
	int pos; // position
	enum DATETYPE { RFC1123, RFC850, ASC } dtType;
	char szarrWeekday[7][16] = {
		"Monday",
		"Tuesday",
		"Wednesday",
		"Thursday",
		"Friday",
		"Saturday",
		"Sunday"
	};
	char szarrWkday[7][8] = {
		"Mon",
		"Tue",
		"Wed",
		"Thu",
		"Fri",
		"Sat",
		"Sun"
	};
	ASSERT(szRFC1123 != NULL && szHTTPDate != NULL);

	if ((noc = ExtractWord2(szWord, 16, szHTTPDate, 0)) <= 0) {
		// Failed to extract a word.
		return false;
	}
	pos = noc;
	i = strlen(szWord);
	if (i >= 1) {
		i--; // Point to the last character.
		if (szWord[i] == ',') {
			// RFC1123 or RFC850
			szWord[i] = '\0'; // Delete the comma.
			dtType = RFC850;
			for (i = 0; i < 7; i++) {
				if (strcmp(szWord, szarrWkday[i]) == 0) {
					// RFC1123
					dtType = RFC1123;
					break;
				}
			}
			// If no match, the date type remains as RFC850.
		} else {
			// ASCTIME
			dtType = ASC;
		}
	} else {
		// i == 0
		return false;
	}
	// szWord contains the weekday or wkday.
	// pos points to the head of the next word.
	switch (dtType) {
	case RFC1123:
		// Just copy it.
		strncpy(szRFC1123, szHTTPDate, SZGMT_SIZE);
		szRFC1123[SZGMT_SIZE - 1] = '\0';
		break;
	case RFC850:
		{
			char szWkday[10] = "";
			char szDay[3] = "";
			char szMonth[4] = "";
			char szYear[3] = "";
			char szHundYear[3] = "";
			char szTime[9] = "";
			int j;
			// Convert from rfc850 date format to rfc1123 date format.
			// szWord contains the weekday.
			for (i = 0; i < 7; i++) {
				if (strcmp(szWord, szarrWeekday[i]) == 0) {
					break;
				}
			}
			if (i == 7) {
				// No match
				return false;
			}
			strcpy(szWkday, szarrWkday[i]);
			if ((noc = ExtractWord2(szWord, 16, szHTTPDate, pos)) <= 0) {
				return false;
			}
			pos += noc;
			// Get the day.
			// szWord[2] should be the first hyphen.
			for (i = 0, j = 0; i < 2 && szWord[i]; i++, j++) {
				szDay[j] = szWord[i];
			}
			if (i != 2) {
				return false;
			}
			if (szWord[i] != '-') {
				return false;
			}
			szDay[j] = '\0';
			i++;
			// Get the month.
			// szWord[6] should be the second hyphen.
			for (j = 0; i < 6 && szWord[i]; i++, j++) {
				szMonth[j] = szWord[i];
			}
			if (i != 6) {
				return false;
			}
			if (szWord[i] != '-') {
				return false;
			}
			szMonth[j] = '\0';
			i++;
			// Get the year (two digits).
			// szWord[9] should be the end of the word.
			for (j = 0; i < 9 && szWord[i]; i++, j++) {
				szYear[j] = szWord[i];
			}
			if (i != 9) {
				return false;
			}
			if (szWord[i] != '\0') {
				return false;
			}
			szYear[j] = '\0';
			// Assume the year is between 1980 and 2079.
			if (atoi(szYear) >= 80) {
				strcpy(szHundYear, "19");
			} else {
				strcpy(szHundYear, "20");
			}
			// Extract the time.
			// The time always contains 8 characters. The size of the buffer is 9.
			if ((noc = ExtractWord2(szTime, 9, szHTTPDate, pos)) <= 0) {
				return false;
			}
			// Generate the rfc1123 date.
			wsprintf(szRFC1123, _T("%s, %s %s %s%s %s GMT"),
				szWkday,
				szDay,
				szMonth,
				szHundYear, szYear,
				szTime
			);
		}
		break;
	case ASC:
		{
			char szWkday[10] = "";
			char szDay[3] = "";
			char szMonth[4] = "";
			char szYear[5] = ""; // Full year
			char szTime[9] = "";

			// Get the wkday.
			strncpy(szWkday, szWord, 4);
			szWkday[3] = '\0';
			// Get the month.
			if ((noc = ExtractWord2(szMonth, 4, szHTTPDate, pos)) <= 0) {
				return false;
			}
			pos += noc;
			if ((noc = ExtractWord2(szDay, 3, szHTTPDate, pos)) <= 0) {
				return false;
			}
			pos += noc;
			if ((noc = ExtractWord2(szTime, 9, szHTTPDate, pos)) <= 0) {
				return false;
			}
			pos += noc;
			if ((noc = ExtractWord2(szYear, 5, szHTTPDate, pos)) <= 0) {
				return false;
			}
			wsprintf(szRFC1123, "%s, %02d %s %s %s GMT",
				szWkday,
				atoi(szDay),
				szMonth,
				szYear,
				szTime
			);
		}
		break;
	default:
		ASSERT(false);
		break;
	}
	return true;
}
