// ServerThread.h: interface for the CServerThread class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_SERVERTHREAD_H__843A6A82_954D_459E_969A_53002D0962CB__INCLUDED_)
#define AFX_SERVERTHREAD_H__843A6A82_954D_459E_969A_53002D0962CB__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <afxmt.h>
#include <stdlib.h>
#include "Base64.h"

#define SRECEIVEBUFFER_SIZE     200
#define SREQUESTBUFFER_SIZE     2048

#define RQ_SUCCESSFUL           0
#define RQ_ERROR                1
#define RQ_INCOMPLETE           2
#define RQ_SERVER_ERROR         3

#ifndef SD_RECEIVE
#define SD_RECEIVE      0x00
#define SD_SEND         0x01
#define SD_BOTH         0x02
#endif

#define CP_AFTER        0x00
#define CP_BEFORE       0x01

#define MAX_URI             1024
#define SENDBUFFER_SIZE	    4096

#define SZGMT_SIZE          40
#define SZCONTENTTYPE_SIZE  40
#define SZEXTENSION_SIZE    30
#define MAX_USERNAME        40
#define MAX_PASSWORD        40

struct STHREAD_INFO
{
	CString sBasePath; // shared constant
	SOCKET socketServer; // shared thread-safe variable
	CString *sarrPasswords; // shared constant
	int iPasswordCount; // shared constant
	CListBox *plistMonitor; // shared variable
	CString sDefaultPage; // shared constant
	int iThreadCount; // shared variable
	HWND hwndDialog; // shared constant
	bool fToStop; // shared variable
	BOOL *pfRecToMon; // shared variable
	CCriticalSection csMain; // critical section (mutex)
};

class CServerThread  
{
protected:
	CServerThread(); // No object of this type can be created.
public:
	static UINT EntryProcedure(LPVOID lpParam);
	virtual ~CServerThread();
	struct tagTreatRequestState;
	enum RequestMethodConstants;
	enum TRStateConstants
	{
		trHead,
		trBody
	};
	enum GetStateConstants
	{
		gtURIProtocol,
		gtBeforeHeader,
		gtRequestHeader,
		gtAccept,
		gtAcceptCharset,
		gtAcceptEncoding,
		gtAcceptLanguage,
		gtHost,
		gtReferer,
		gtUserAgent,
		gtConnection,
		gtIfModifiedSince,
		gtAuthorization
	};
	struct tagTreatRequestState
	{
		SOCKET socketClient; // The client socket
		SOCKET socketServer; // The server socket
		enum TRStateConstants state; // The state of treatment to the request
		enum GetStateConstants getState; // The state of treatment to the GET and HEAD request
		enum RequestMethodConstants iRMethod; // Request method (GET, HEAD, OPTION, etc.)
		char *szURI; // The received URI (Universal Resource Identifier)
		char *szPath; // The path converted from the URI
		char *sSendBuffer; // Buffer for sending data
		const char *szBasePath; // The base path (the WWW root directory)
		bool fKeepAlive; // Whether the server required to keep alive
		bool fRequestHead; // Whether the request is HEAD (to differ from GET)
		bool fIMS; // Whether the server sent a If-Modified-Since GMT
		char szIMS[SZGMT_SIZE]; // String for storing If-Modified-Since time information
		char *szFullPath; // The full path converted from the path and the base path
		char szUserName[MAX_USERNAME]; // The user name sent by the server
		char szPassword[MAX_PASSWORD]; // The password sent by the server
		CString *sarrPasswords; // The array of passwords
		int iPasswordCount; // The number of passwords
		CListBox *plistMonitor; // Pointer to the monitor list box
		const char *szDefaultPage;
		BOOL *pfRecToMon;
		CCriticalSection *pcsMain; // pointer to the critical section
		HWND hwndDialog; // handle to the dialog
	};
	enum RequestMethodConstants
	{
		rmUnknown,
		rmOptions,
		rmGet,
		rmHead,
		rmPost,
		rmPut,
		rmDelete,
		rmTrace,
		rmConnect,
		rmIncomplete
	};
protected:
	static int TreatRequest_HasALine2(const char *sRequest, int *piBegin, int iEnd);
	static bool ShutSocketDown(SOCKET socketOne);
	static bool ExtractUserPass(char *szUserName, int iSizeUser, char *szPassword, int iSizePass, const char *szUserPass);
	static bool Authorize(
		const char *szFullPath, const char *szBasePath, 
		const char *szUserName, const char *szPassword, CString *sarrPasswords, 
		int iPasswordCount, CListBox *plistMonitor, BOOL *pfRecToMon,
		HWND hwndDialog
		);
	static bool SendHTTPError(int iError, SOCKET socketClient);
	static bool IsValidFile(const char *szFullPath, const char *szBasePath);
	static int ConvertURIToPath(char *szPath, int iSize, const char *szURI);
	static bool TreatRequest_HasALine(const char *sRequest, int iBegin, int iEnd);
	static int IgnoreLine(const char *sRequest, int *piBegin, int iEnd);
	static void ShowRequest(const char* szRequest, int iBegin, int iEnd);
	static int TreatRequestGet(struct tagTreatRequestState *pTRState, const char *szRequest, int *piBegin, int iEnd);
	static CServerThread::RequestMethodConstants AnalyzeRequestMethod(const char *szRequest, int *piBegin, int iEnd);
	static int TreatRequest(struct tagTreatRequestState *pTRState, const char *szRequest, int *piBegin, int iEnd);
	static const int s_iTimeoutLimit;
};

int ExtractWord(char *szWord, int iSize, const char *szRequest, int iBegin, int iEnd);
int OldExtractLine(char *szLine, int iSize, const char *sRequest, int iBegin, int iEnd);
int DiscardSpace(const char *sRequest, int *piBegin, int iEnd);
int ExtractLine(char *szLine, int iSize, const char *sRequest, int iBegin, int iEnd, int *piBytesExtracted);
int ExtractWord2(TCHAR *szWord, int iSize, const TCHAR *szLine, int iBegin);
char hdtochar(char cHexDigitHigh, char cHexDigitLow);
char *CombinePath(TCHAR *szPath, int iSize, const TCHAR *szPathOther, int iMethod);
char *SystemTimeToStringGMT(char *szGMT, const SYSTEMTIME *psystimeOne);
bool ExtractExtension(TCHAR *szExtension, int iSize, const TCHAR *szPath);
bool ExtensionToContentType(char *szContentType, int iSize, const char *szExtension);
bool ConvertToFullPath(char *szFullPath, int iSize, const char *szPath);
bool HTTPDateTo1123(char *szRFC1123, const char *szHTTPDate);

#endif // !defined(AFX_SERVERTHREAD_H__843A6A82_954D_459E_969A_53002D0962CB__INCLUDED_)
